import bmcs_utils.api as bu
import traits.api as tr
from bmcs_utils.api import mpl_align_yaxis
import numpy as np
from scipy.integrate import cumtrapz

class InelStateEvolution(bu.InteractiveModel):
    name = 'State evolution'

    slider_exp = tr.WeakRef(bu.InteractiveModel)

    t_slider = bu.Float(0)
    t_max = bu.Float(1.001)

    t_arr = tr.DelegatesTo('slider_exp')
    Sig_arr = tr.DelegatesTo('slider_exp')
    Eps_arr = tr.DelegatesTo('slider_exp')
    s_x_t = tr.DelegatesTo('slider_exp')
    s_y_t = tr.DelegatesTo('slider_exp')
    w_t = tr.DelegatesTo('slider_exp')
    iter_t = tr.DelegatesTo('slider_exp')

    ipw_view = bu.View(
        bu.Item('t_slider', latex=r't',
                editor=bu.FloatRangeEditor(low=0, high_name='t_max', n_steps=50)),
        bu.Item('t_max', latex=r't_{\max}'),
    )

    def plot_Sig_Eps(self, axes):
        ax1, ax11, ax2, ax22, ax3, ax33, ax4, ax44 = axes
        colors = ['blue', 'red', 'green', 'black', 'magenta']
        t = self.t_arr
        s_x_pi_, s_y_pi_, w_pi_, z_, alpha_x_, alpha_y_, omega_s_, omega_w_ = self.Eps_arr.T
        tau_x_pi_, tau_y_pi_, sig_pi_, Z_, X_x_, X_y_, Y_s_, Y_w_ = self.Sig_arr.T
        n_step = len(s_x_pi_)

        idx = np.argmax(self.t_slider < self.t_arr)

        # slip path in 2d
        def get_cum_s(s_x, s_y):
            d_s_x, d_s_y = s_x[1:] - s_x[:-1], s_y[1:] - s_y[:-1]
            d_s = np.hstack([0, np.sqrt(d_s_x**2 + d_s_y**2)])
            return cumtrapz(d_s, initial=0)

        s_t = get_cum_s(self.s_x_t, self.s_y_t)
        s_pi_t = get_cum_s(s_x_pi_, s_y_pi_)
        w_t = self.w_t
        w_pi_t = w_pi_
        tau_pi = np.sqrt(tau_x_pi_**2 + tau_y_pi_**2)

        ax1.set_title('stress - displacement');
        ax1.plot(t, tau_pi, '--', color='darkgreen', label=r'$||\tau||$')
        ax1.fill_between(t, tau_pi, 0, color='limegreen', alpha=0.1)
        ax1.plot(t, sig_pi_, '--', color='olivedrab', label = r'$\sigma$')
        ax1.fill_between(t, sig_pi_, 0, color='olivedrab', alpha=0.1)
        ax1.set_ylabel(r'$|| \tau ||, \sigma$')
        ax1.set_xlabel('$t$');
        ax1.plot(t[idx], 0, marker='H', color='red')
        ax1.legend()
        ax11.plot(t, s_t, color='darkgreen', label=r'$||s||$')
        ax11.plot(t, s_pi_t, '--', color='orange', label=r'$||s^\pi||$')
        ax11.plot(t, w_t, color='olivedrab', label=r'$w$')
        ax11.plot(t, w_pi_t, '--', color='chocolate', label=r'$w^\pi$')
        ax11.set_ylabel(r'$|| s ||, w$')
        ax11.legend()
        mpl_align_yaxis(ax1,0,ax11,0)

        ax2.set_title('energy release rate - damage');
        ax2.plot(t, Y_w_, '--', color='darkgray', label=r'$Y_w$')
        ax2.fill_between(t, Y_w_, 0, color='darkgray', alpha=0.15)
        ax2.plot(t, Y_s_, '--', color='darkslategray', label=r'$Y_s$')
        ax2.fill_between(t, Y_s_, 0, color='darkslategray', alpha=0.05)
        ax2.set_xlabel('$t$');
        ax2.set_ylabel('$Y$')
        ax2.plot(t[idx], 0, marker='H', color='red')
        ax2.legend()
        ax22.plot(t, omega_w_, color='darkgray', label=r'$\omega_w$')
        ax22.plot(t, omega_s_, color='darkslategray', label=r'$\omega_s$')
        ax22.set_ylim(ymax=1)
        ax22.set_ylabel(r'$\omega$')
        ax22.legend()

        ax3.set_title('hardening force - displacement');
        alpha_t = np.sqrt(alpha_x_**2 + alpha_y_**2)
        X_t = np.sqrt(X_x_**2 + X_y_**2)
        ax3.plot(t, Z_, '--', color='darkcyan', label=r'$Z$')
        ax3.fill_between(t, Z_, 0, color='darkcyan', alpha=0.05)
        ax3.plot(t, X_t, '--', color='darkslateblue', label=r'$X$')
        ax3.fill_between(t, X_t, 0, color='darkslateblue', alpha=0.05)
        ax3.set_ylabel(r'$Z, X$')
        ax3.set_xlabel('$t$');
        ax3.plot(t[idx], 0, marker='H', color='red')
        ax3.legend()
        ax33.plot(t, z_, color='darkcyan', label=r'$z$')
        ax33.plot(t, alpha_t, color='darkslateblue', label=r'$\alpha$')
        ax33.set_ylabel(r'$z, \alpha$')
        ax33.legend(loc='lower left')

        slide_model = self.slider_exp.slide_model
        slide_model.plot_f_state(ax4, self.Eps_arr[idx,:], self.Sig_arr[idx,:] )

    @staticmethod
    def subplots(fig):
        ((ax1, ax2), (ax3, ax4)) = fig.subplots(2, 2)
        ax11 = ax1.twinx()
        ax22 = ax2.twinx()
        ax33 = ax3.twinx()
        ax44 = ax4.twinx()
        return ax1, ax11, ax2, ax22, ax3, ax33, ax4, ax44

    def update_plot(self, axes):
        self.plot_Sig_Eps(axes)



#!/usr/bin/env python
# coding: utf-8

# # Derivation of the level set function for SLIDE interface model 3.2

# In[32]:


import sympy as sp
import numpy as np

# ![image.png](attachment:image.png)

# # Variables
# 
# **Stress variables**
# 
# $x$ corresponds to $\sigma_N$ and $y$ corresponds to $\sigma_T$

# In[33]:


x, y = sp.symbols('x, y')


# **Unknown parameters of the ellipse cap function**

# In[34]:

x_c = sp.symbols('x_c') # center point of an elips
a = sp.symbols('a', nonnegative=True)
b = sp.symbols('b', nonnegative=True)
c = sp.symbols('c', positive=True)

# **Material parameters**

# In[35]:

x_0 = sp.symbols('x_0')
x_bar, y_bar = sp.symbols(r'x_bar, y_bar', nonnegative=True )
m = sp.symbols('m')

# # Mid part of the threshold function

# In[36]:

f_mid_ = sp.sqrt( y**2 ) - (y_bar - m * (x-x_0))

# # Cap function
# 
# It is introduced in form of an ellipse centered at the position ($x_c, 0$)

f_cap_ = sp.sqrt( (x-x_0-x_c)**2/a**2 + y**2/b**2 ) - c

# Construct the derivatives

df_cap_dx_ = f_cap_.diff(x)
df_cap_dy_ = f_cap_.diff(y)

# # Value and continuity conditions specified in the figure above

eq1 = sp.Eq( f_cap_.subs({x:x_bar, y:0}), 0 )
eq2 = sp.Eq( f_cap_.subs({x:x_0, y:y_bar}), 0)
eq3 = sp.Eq( ( -df_cap_dx_ / df_cap_dy_).subs({x:x_0, y:y_bar}), -m)

# Solve for $a, b$ and $x_c$

abx_subs = sp.solve({eq1,eq2,eq3},{a,b,x_c})[0]
abx_subs

# # Continuity between $f_\mathrm{mid}$ and $f_\mathrm{cap}$
# 

# **Require an identical value for**
# \begin{align}
# f_\mathrm{mid}(x_c, 0) = f_\mathrm{cap}(x_c,0)
# \end{align}

f_mid_c_ = f_mid_.subs({x:abx_subs[x_c]+x_0,y:0})
f_cap_abx_ = f_cap_.subs(abx_subs)
f_cap_abxc_ = f_cap_abx_.subs({x:abx_subs[x_c]+x_0,y:0})
eq4 = sp.Eq( f_cap_abxc_, f_mid_c_ )

# Which can be used to express $c$

c_solved = sp.solve(eq4, c)[0]

# Substitute back the $f_\mathrm{cap}$ to obtain its resolved form

f_cap_solved_ = f_cap_abx_.subs(c, c_solved)

# Test the obtained position of $x_c$
x_c_ = abx_subs[x_c]

# # Domain separation for $f_\mathrm{cap}$ and $f_\mathrm{mid}$
# 
# Define the transition between cap and mid domains by defining a connection 
# line between [$x_c$,0]-[0,$\bar{\tau}$]. 

y_trans_ = (-y_bar / (x_c) * (x - x_0 -x_c)).subs(x_c, x_c_)
f_cap_domain_ = sp.sign(x_bar-x_0) * sp.sign(-m) * (sp.Abs(y) - y_trans_)  > 0

# # Visualization

f_t, f_c, f_c0, tau_bar = sp.symbols('f_t, f_c, f_c0, tau_bar')
subs_tension = {x_0:0, x_bar:f_t, y_bar:tau_bar}
subs_shear = {y_bar:tau_bar, x_0:0}
subs_compression = {x_0: -f_c0, x_bar:-f_c,  y_bar: tau_bar-m*(-f_c0) }

import bmcs_utils.api as bu

class FDoubleCapExpr(bu.SymbExpr):
    # -------------------------------------------------------------------------
    # Symbolic derivation of variables
    # -------------------------------------------------------------------------

    x = x

    y = y

    # -------------------------------------------------------------------------
    # Model parameters
    # -------------------------------------------------------------------------

    f_t, f_c, f_c0, tau_bar, m = f_t, f_c, f_c0,tau_bar, m

    # -------------------------------------------------------------------------
    # Expressions
    # -------------------------------------------------------------------------

    f_solved = sp.Piecewise(
        (f_cap_solved_.subs(subs_tension), f_cap_domain_.subs(subs_tension)),
        (f_cap_solved_.subs(subs_compression), f_cap_domain_.subs(subs_compression)),
        (f_mid_.subs(subs_shear), True)
    )

    #-------------------------------------------------------------------------
    # Declaration of the lambdified methods
    #-------------------------------------------------------------------------

    symb_model_params = ['f_t', 'f_c', 'f_c0', 'tau_bar', 'm']

    # List of expressions for which the methods `get_`
    symb_expressions = [
        ('f_solved', ('x', 'y')),
    ]

max_f_c = 100
max_f_t = 10
max_tau_bar = 10

class FDoubleCap(bu.InteractiveModel,bu.InjectSymbExpr):

    name = 'Threshold'

    symb_class = FDoubleCapExpr

    f_t = bu.Float(5, MAT=True)
    f_c = bu.Float(80, MAT=True)
    f_c0 = bu.Float(30, MAT=True)
    tau_bar = bu.Float(5, MAT=True)
    m = bu.Float(0.1, MAT=True)

    ipw_view = bu.View(
        bu.Item('f_t', editor=bu.FloatRangeEditor(low=1, high=max_f_t)),
        bu.Item('f_c', editor=bu.FloatRangeEditor(low=10, high=max_f_c)),
        bu.Item('f_c0', latex='f_{c0}', editor=bu.FloatRangeEditor(low=5, high=0.9*max_f_c)),
        bu.Item('tau_bar', latex=r'\bar{\tau}', editor=bu.FloatRangeEditor(low=1, high=max_tau_bar)),
        bu.Item('m', minmax=(0.0001, 0.5))
    )

    def subplots(self, fig):
        ax = fig.subplots(1, 1)
#        ax = fig.add_subplot(1, 1, 1, projection='3d')
        return ax

    def plot_3d(self, ax):
        X_a, Y_a = np.mgrid[-1.1*max_f_c:1.1*max_f_t:210j, -max_tau_bar:max_tau_bar:210j]
        Z_a = self.symb.get_f_solved(X_a, Y_a)
        #ax.contour(X_a, Y_a, Z_a, levels=8)
        Z_0 = np.zeros_like(Z_a)
        ax.plot_surface(X_a, Y_a, Z_a, rstride=1, cstride=1,
                        cmap='winter', edgecolor='none')
        ax.plot_surface(X_a, Y_a, Z_0, edgecolor='none')
        ax.set_title('threshold function');

    def plot_contour(self, ax):
        X_a, Y_a = np.mgrid[-1.05*max_f_c:1.1*max_f_t:210j, -max_tau_bar:max_tau_bar:210j]
        Z_a = self.symb.get_f_solved(X_a, Y_a)
        ax.contour(X_a, Y_a, Z_a, levels=0)
        ax.set_title('threshold function');


    def update_plot(self, ax):
        # Evaluate the threshold function within an orthogonal grid
        self.plot_contour(ax)


import os

import matplotlib

from bmcs_matmod.ms1.vmats3D_mpl_csd_eeq import MATS3DMplCSDEEQ
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

DELTA = np.identity(3)

EPS = np.zeros((3, 3, 3), dtype='f')
EPS[(0, 1, 2), (1, 2, 0), (2, 0, 1)] = 1
EPS[(2, 1, 0), (1, 0, 2), (0, 2, 1)] = -1

DD = np.hstack([DELTA, np.zeros_like(DELTA)])
EEPS = np.hstack([np.zeros_like(EPS), EPS])

GAMMA = np.einsum(
    'ik,jk->kij', DD, DD
) + np.einsum(
    'ikj->kij', np.fabs(EEPS)
)

def get_eps_ab(eps_O): return np.einsum(
    'Oab,...O->...ab', GAMMA, eps_O
)[np.newaxis, ...]

GAMMA_inv = np.einsum(
    'aO,bO->Oab', DD, DD
) + 0.5 * np.einsum(
    'aOb->Oab', np.fabs(EEPS)
)


def get_sig_O(sig_ab): return np.einsum(
    'Oab,...ab->...O', GAMMA_inv, sig_ab
)[0, ...]


GG = np.einsum(
    'Oab,Pcd->OPabcd', GAMMA_inv, GAMMA_inv
)


def get_K_OP(D_abcd):
    return np.einsum(
        'OPabcd,abcd->OP', GG, D_abcd
    )

# The above operators provide the three mappings
# map the primary variable to from vector to field
# map the residuum field to evctor (assembly operator)
# map the gradient of the residuum field to system matrix


def get_UF_t(mats, time_function, n_t):

    # @todo: put into the explorer  - rename int_vars to state_vars
    int_var = np.zeros((n_mp, 25))
    int_var_aux = np.zeros((n_mp, 25))
    dissip = np.zeros((n_mp, 8))
    save = np.concatenate((int_var, dissip), axis=1)
    df = pd.DataFrame(save)
    df.to_hdf(path, 'first', mode='w', format='table')
    D = np.zeros((3, 3, 3, 3))
    D = D[np.newaxis, :, :, :, :]

    # total number of DOFs
    n_O = 6
    # Global vectors
    F_ext = np.zeros((n_O,), np.float_)
    F_O = np.zeros((n_O,), np.float_)
    U_P = np.zeros((n_O,), np.float_)
    U_k_O = np.zeros((n_O,), dtype=np.float_)
    eps_aux = get_eps_ab(U_k_O)
    # Construct index maps distinguishing the controlled displacements
    # and free displacements. Here we simply say the the control displacement
    # is the first one. Then index maps are constructed using the np.where
    # function returning the indexes of True positions in a logical array.
    CONTROL = 0
    FREE = slice(1, None)  # This means all except the first index, i.e. [1:]
    # Setup the system matrix with displacement constraints
    # Time stepping parameters
    t_n1, t_max, t_step = 0, len(time_function), 1 / n_t
    t_n = 0
    # Iteration parameters
    k_max, R_acc = 1000, 1e-3
    # Record solutions
    U_t_list, F_t_list, U_P_list = [np.copy(U_k_O)], [np.copy(F_O)], [np.copy(U_P)]

    # Load increment loop
    while t_n1 <= t_max - 1:
        #print('t:', t_n1)
        # Get the displacement increment for this step
        delta_U = time_function[t_n1] - time_function[t_n]
        k = 0
        # Equilibrium iteration loop
        while k < k_max:
            # Transform the primary vector to field
            eps_ab = get_eps_ab(U_k_O).reshape(3, 3)
            # Stress and material stiffness
            D_abcd, sig_ab, eps_p_Emab, sig_ab_int = mats.get_corr_pred(
                eps_ab, 1, int_var, eps_aux, F_ext
            )
            # Internal force
            F_O = get_sig_O(sig_ab.reshape(1, 3, 3)).reshape(6, )

            # System matrix
            K_OP = get_K_OP(D_abcd)
            #Beta = get_K_OP(beta_Emabcd)
            # Get the balancing forces - NOTE - for more displacements
            # this should be an assembly operator.
            # KU remains a 2-d array so we have to make it a vector
            KU = K_OP[:, CONTROL] * delta_U
            # Residuum
            R_O = F_ext - F_O - KU
            # Convergence criterion
            R_norm = np.linalg.norm(R_O[FREE])
            if R_norm < R_acc:
                # Convergence reached
                break
            # Next iteration -
            delta_U_O = np.linalg.solve(K_OP[FREE, FREE], R_O[FREE])
            # Update total displacement
            U_k_O[FREE] += delta_U_O
            # Update control displacement
            U_k_O[CONTROL] += delta_U
            # Note - control displacement nonzero only in the first iteration.
            delta_U = 0
            k += 1
        else:
            print('no convergence')
            break

        # Target time of the next load increment
#         U_t_list.append(np.copy(U_k_O))
#         F_t_list.append(F_O)

            # Update states variables after convergence
        int_var = mats._get_state_variables(eps_ab, int_var, eps_aux)

        # Definition internal variables / forces per column:  1) damage N, 2)iso N, 3)kin N, 4) consolidation N, 5) eps p N,
        # 6) sigma N, 7) iso F N, 8) kin F N, 9) energy release N, 10) damage T, 11) iso T, 12-14) kin T, 15-17) eps p T,
        # 18-20) sigma T, 21) iso F T, 22-24) kin F T, 25) energy release T

        # Definition dissipation components per column: 1) damage N, 2) damage T, 3) eps p N, 4) eps p T, 5) iso N
        # 6) iso T, 7) kin N, 8) kin T

        dissip[:, 0] += np.einsum('...n,...n->...n', int_var[:, 0] - int_var_aux[:, 0], int_var[:, 8])
        dissip[:, 1] += np.einsum('...n,...n->...n', int_var[:, 9] - int_var_aux[:, 9], int_var[:, 20])
        dissip[:, 2] += np.einsum('...n,...n->...n', int_var[:, 4] - int_var_aux[:, 4], int_var[:, 5])
        dissip[:, 3] += np.einsum('...n,...n->...', int_var[:, 14:17] - int_var_aux[:, 14:17], int_var[:, 17:20])
        dissip[:, 4] += np.einsum('...n,...n->...n', int_var[:, 1] - int_var_aux[:, 1], int_var[:, 6])
        dissip[:, 5] += np.einsum('...n,...n->...n', int_var[:, 10] - int_var_aux[:, 10], int_var[:, 17])
        dissip[:, 6] += np.einsum('...n,...n->...n', int_var[:, 2] - int_var_aux[:, 2], int_var[:, 7])
        dissip[:, 7] += np.einsum('...n,...n->...', int_var[:, 11:14] - int_var_aux[:, 11:14], int_var[:, 21:24])

        int_var_aux = int_var * 1

        save = np.concatenate((int_var, dissip), axis=1)

        df = pd.DataFrame(save)
        df.to_hdf(path, 'middle' + np.str(t_n1), append=True)

        U_t_list.append(np.copy(U_k_O))
        F_t_list.append(F_O)
        U_P_list.append(U_P)
        eps_aux = get_eps_ab(U_k_O)
        D_aux = D_abcd[np.newaxis, :, :, :, :]
        D = np.concatenate((D, D_aux))

        t_n = t_n1
        t_n1 += 1

    U_t, F_t, U_p = np.array(U_t_list), np.array(F_t_list), np.array(U_P_list)
    return U_t, F_t, D, U_p

def get_int_var(path, size, n_mp):  # unpacks saved data

    S = np.zeros((len(F), n_mp, 33))

    S[0] = np.array(pd.read_hdf(path, 'first'))

    for i in range(1, size):
        S[i] = np.array(pd.read_hdf(path, 'middle' + np.str(i - 1)))

    # @todo: definition of protocols - order should not be defined by indexing
    omega_N_Emn = S[:, :, 0]
    z_N_Emn = S[:, :, 1]
    alpha_N_Emn = S[:, :, 2]
    r_N_Emn = S[:, :, 3]
    eps_N_p_Emn = S[:, :, 4]
    sigma_N_Emn = S[:, :, 5]
    Z_N_Emn = S[:, :, 6]
    X_N_Emn = S[:, :, 7]
    Y_N_Emn = S[:, :, 8]

    omega_T_Emn = S[:, :, 9]
    z_T_Emn = S[:, :, 10]
    alpha_T_Emna = S[:, :, 11:14]
    eps_T_pi_Emna = S[:, :, 14:17]
    sigma_T_Emna = S[:, :, 17:20]
    Z_T_pi_Emn = S[:, :, 20]
    X_T_pi_Emna = S[:, :, 21:24]
    Y_T_pi_Emn = S[:, :, 24]

    Disip_omena_N_Emn = S[:, :, 25]
    Disip_omena_T_Emn = S[:, :, 26]
    Disip_eps_p_N_Emn = S[:, :, 27]
    Disip_eps_p_T_Emn = S[:, :, 28]
    Disip_iso_N_Emn = S[:, :, 29]
    Disip_iso_T_Emn = S[:, :, 30]
    Disip_kin_N_Emn = S[:, :, 31]
    Disip_kin_T_Emn = S[:, :, 32]

    return omega_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn, Z_N_Emn, X_N_Emn, Y_N_Emn, \
           omega_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_T_Emna, Z_T_pi_Emn, X_T_pi_Emna, Y_T_pi_Emn, \
           Disip_omena_N_Emn, Disip_omena_T_Emn, Disip_eps_p_N_Emn, Disip_eps_p_T_Emn, Disip_iso_N_Emn, \
           Disip_iso_T_Emn, Disip_kin_N_Emn, Disip_kin_T_Emn

concrete_type= 0        # 0:C40MA, 1:C80MA, 2:120MA, 3:Tensile, 4:Compressive, 5:Biaxial

# @todo: convert this to a list of dictionaries
Concrete_Type_string = [
    'C40MA', 'C80MA','C120MA',
    'Tensile', 'Compressive', 'Biaxial'
]

# @todo: convert to a load scenario class
loading_scenario = 'monotonic'   # monotonic, cyclic

M_plot = 0  # Plot microplanes polar graphs. 1: yes, 0: no

t_steps = 500
n_mp = 28

if loading_scenario == 'monotonic':

    eps = -0.01

    load = np.linspace(0, eps, t_steps)

if loading_scenario == 'cyclic':

    eps = [-0.003, -0.00117, -0.0053, -0.002, -0.0065, -0.002, -0.0085, -0.003, -0.01]

    load1 = np.linspace(0, eps[0], t_steps)
    load2 = np.linspace(load1[-1], eps[1], t_steps)
    load3 = np.linspace(load2[-1], eps[2], t_steps)
    load4 = np.linspace(load3[-1], eps[3], t_steps)
    load5 = np.linspace(load4[-1], eps[4], t_steps)
    load6 = np.linspace(load5[-1], eps[5], t_steps)
    load7 = np.linspace(load6[-1], eps[6], t_steps)
    load8 = np.linspace(load7[-1], eps[7], t_steps)
    load9 = np.linspace(load8[-1], eps[8], t_steps)

    load = np.concatenate((load1, load2[1::], load3[1::], load4[1::], load5[1::], load6[1::], load7[1::], load8[1::], load9[1::]))

t_steps = len(load)

home_dir = os.path.expanduser('~')

if not os.path.exists('Data Processing'):
    os.makedirs('Data Processing')

path = os.path.join(
   home_dir, 'Data Processing',
   '3Ddc' + Concrete_Type_string[concrete_type] + loading_scenario + '.hdf5')

ms1 = MATS3DMplCSDEEQ(concrete_type)

U, F, D, U_p = get_UF_t(
    ms1,
    load,
    t_steps
)

[omega_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn, Z_N_Emn, X_N_Emn, Y_N_Emn, omega_T_Emn, z_T_Emn,
 alpha_T_Emna, eps_T_pi_Emna, sigma_T_Emna, Z_T_pi_Emn, X_T_pi_Emna, Y_T_pi_Emn, Disip_omena_N_Emn, Disip_omena_T_Emn,
 Disip_eps_p_N_Emn, Disip_eps_p_T_Emn, Disip_iso_N_Emn, Disip_iso_T_Emn, Disip_kin_N_Emn, Disip_kin_T_Emn] \
    = get_int_var(path, len(F), n_mp)

font = {'family': 'DejaVu Sans',
        'size': 18}

matplotlib.rc('font', **font)

print(np.max(np.abs(F[:, 0])), 'fc')

f, (ax2) = plt.subplots(1, 1, figsize=(15, 12))

ax2.plot(np.abs(U[:, 0]), np.abs(F[:, 0]), 'k', linewidth=3.5)
ax2.set_xlabel(r'$|\varepsilon_{11}$| [-]', fontsize=25)
ax2.set_ylabel(r'$|\sigma{11}$| [-]', fontsize=25)
#ax2.set_ylim(0,70)

print(np.max(np.abs(F[:, 0])), 'fc')

eps_T_pi_Emn = np.sqrt(np.einsum('...i,...i->... ', eps_T_pi_Emna, eps_T_pi_Emna))
sigma_T_Emn = np.sqrt(np.einsum('...i,...i->... ', sigma_T_Emna, sigma_T_Emna))
X_T_pi_Emn = np.sqrt(np.einsum('...i,...i->... ', X_T_pi_Emna, X_T_pi_Emna))
alpha_T_Emn = np.sqrt(np.einsum('...i,...i->... ', alpha_T_Emna, alpha_T_Emna))

plt.figure(figsize=(27, 12))

plt.subplot(241)
for i in range(n_mp):
    plt.plot(np.abs(U[:, 0]), omega_N_Emn[:, i])
plt.title(r'$\omega_N$')

plt.subplot(242)
for i in range(n_mp):
    plt.plot(np.abs(U[:, 0]), eps_N_p_Emn[:, i])
plt.title(r'$\varepsilon^p_N$')

plt.subplot(243)
for i in range(n_mp):
    plt.plot(np.abs(U[:, 0]), Z_N_Emn[:, i])
plt.title(r'$Z_N$')

plt.subplot(244)
for i in range(n_mp):
    plt.plot(np.abs(U[:, 0]), X_N_Emn[:, i])
plt.title(r'$X_N$')

plt.subplot(245)
for i in range(n_mp):
    plt.plot(np.abs(U[:, 0]), omega_T_Emn[:, i])
plt.title(r'$\omega_T$')

plt.subplot(246)
for i in range(n_mp):
    plt.plot(np.abs(U[:, 0]), eps_T_pi_Emn[:, i])
plt.title(r'$\varepsilon^p_T$')

plt.subplot(247)
for i in range(n_mp):
    plt.plot(np.abs(U[:, 0]), Z_T_pi_Emn[:, i])
plt.title(r'$Z_T$')

plt.subplot(248)
for i in range(n_mp):
    plt.plot(np.abs(U[:, 0]), X_T_pi_Emn[:, i])
plt.title(r'$X_T$')

plt.show()



import bmcs_utils.api as bu
import numpy as np
import traits.api as tr
from bmcs_matmod.slide.slide_32 import Slide32, ConvergenceError
from bmcs_matmod.slide.energy_dissipation import EnergyDissipation
from bmcs_matmod.slide.inel_state_evolution import InelStateEvolution
from bmcs_matmod.time_fn.time_function import TimeFunction

class SlideExplorer(bu.InteractiveModel):
    name = 'Explorer'

    slide_model = tr.Instance(Slide32, ())

    energy_dissipation = tr.Instance(EnergyDissipation)
    '''Viewer to the energy dissipation'''
    def _energy_dissipation_default(self):
        return EnergyDissipation(slider_exp=self)

    inel_state_evolution = tr.Instance(InelStateEvolution)
    '''Viewer to the inelastic state evolution'''
    def _inel_state_evolution_default(self):
        return InelStateEvolution(slider_exp=self)

    time_fn = tr.Instance(TimeFunction, ())

    def __init__(self, *args, **kw):
        super(SlideExplorer, self).__init__(*args, **kw)
        self.reset_i()

    n_Eps = tr.Property()

    def _get_n_Eps(self):
        return len(self.slide_model.symb.Eps)

    s_x_1 = bu.Float(0, INC=True)
    s_y_1 = bu.Float(0, INC=True)
    w_1 = bu.Float(0, INC=True)

    n_steps = bu.Int(10, ALG=True)
    k_max = bu.Int(20, ALG=True)

    Sig_arr = tr.Array
    Eps_arr = tr.Array

    Sig_t = tr.Property
    def _get_Sig_t(self):
        return self.Sig_arr

    Eps_t = tr.Property
    def _get_Eps_t(self):
        return self.Eps_arr

    ipw_view = bu.View(
        bu.Item('s_x_1', latex=r's_x', minmax=(-4, 4)),
        bu.Item('s_y_1', latex=r's_y', minmax=(-4, 4)),
        bu.Item('w_1', latex=r'w', minmax=(-4, 4)),
        bu.Item('n_steps'),
        bu.Item('k_max'),
        simulator='run',
        reset_simulator='reset'
    )

    def reset_i(self):
        self.s_x_0, self.s_y_0, self.w_0 = 0, 0, 0
        self.t0 = 0
        self.Sig_arr = np.zeros((0, self.n_Eps))
        self.Eps_arr = np.zeros((0, self.n_Eps))
        self.Sig_record = []
        self.Eps_record = []
        self.iter_record = []
        self.t_arr = []
        self.s_x_t, self.s_y_t, self.w_t = [], [], []
        self.Eps_n1 = np.zeros((self.n_Eps,), dtype=np.float_)
        self.s_x_1 = 0
        self.s_y_1 = 0
        self.w_1 = 0

    def get_response_i(self, update_progress=lambda t: t):
        # global Eps_record, Sig_record, iter_record
        # global t_arr, s_x_t, s_y_t, w_t, s_x_0, s_y_0, w_0, t0, Eps_n1
        n_steps = self.n_steps
        t_i = np.linspace(0,1, n_steps+1)
        t1 = self.t0 + 1
        ti_arr = np.linspace(self.t0, t1, n_steps + 1)
        si_x_t = np.linspace(self.s_x_0, self.s_x_1, n_steps + 1) + 1e-9
        si_y_t = np.linspace(self.s_y_0, self.s_y_1, n_steps + 1) + 1e-9
        wi_t = np.linspace(self.w_0, self.w_1, n_steps + 1) + 1e-9
        for t, s_x_n1, s_y_n1, w_n1 in zip(t_i, si_x_t, si_y_t, wi_t):
            try: self.Eps_n1, Sig_n1, k = self.slide_model.get_sig_n1(
                    s_x_n1, s_y_n1, w_n1, self.Eps_n1, self.k_max
                )
            except ConvergenceError as e:
                print(e)
                break
            self.Sig_record.append(Sig_n1)
            self.Eps_record.append(self.Eps_n1)
            self.iter_record.append(k)
            update_progress(t)

        self.Sig_arr = np.array(self.Sig_record, dtype=np.float_)
        self.Eps_arr = np.array(self.Eps_record, dtype=np.float_)
        self.iter_t = np.array(self.iter_record, dtype=np.int_)
        n_i = len(self.iter_t)
        self.t_arr = np.hstack([self.t_arr, ti_arr])[:n_i]
        self.s_x_t = np.hstack([self.s_x_t, si_x_t])[:n_i]
        self.s_y_t = np.hstack([self.s_y_t, si_y_t])[:n_i]
        self.w_t = np.hstack([self.w_t, wi_t])[:n_i]
        self.t0 = t1
        self.s_x_0, self.s_y_0, self.w_0 = self.s_x_1, self.s_y_1, self.w_1
        # set the last step index in the response browser
        self.inel_state_evolution.t_max = self.t_arr[-1]
        return

    # ## Plotting functions
    # To simplify postprocessing examples, here are two aggregate plotting functions, one for the state and force variables, the other one for the evaluation of energies

    def plot_sig_w(self, ax):
        sig_t = self.Sig_arr.T[2, ...]
        ax.plot(self.w_t, sig_t, color='orange', lw=3)

    def plot3d_Sig_Eps(self, ax3d):
        tau_x, tau_y = self.Sig_arr.T[:2, ...]
        tau = np.sqrt(tau_x ** 2 + tau_y ** 2)
        ax3d.plot3D(self.s_x_t, self.s_y_t, tau, color='orange', lw=3)

    def subplots(self, fig):
        ax_sxy = fig.add_subplot(1, 2, 1, projection='3d')
        ax_sig = fig.add_subplot(1, 2, 2)

        return ax_sxy, ax_sig

    def run(self, update_progress=lambda t: t):
        try:
            self.get_response_i(update_progress)
        except ValueError:
            print('No convergence reached')
            return

    def reset(self):
        self.reset_i()

    def update_plot(self, axes):
        ax_sxy, ax_sig = axes
        self.plot_sig_w(ax_sig)
        ax_sig.set_xlabel(r'$w$ [mm]');
        ax_sig.set_ylabel(r'$\sigma$ [MPa]');

        #    plot_tau_s(ax1, Eps_arr[-1,...],s_max,500,get_g3,**kw)
        self.plot3d_Sig_Eps(ax_sxy)
        ax_sxy.plot(self.s_x_t, self.s_y_t, 0, color='red')
        ax_sxy.set_xlabel(r'$s_x$ [mm]');
        ax_sxy.set_ylabel(r'$s_y$ [mm]');
        ax_sxy.set_zlabel(r'$\| \tau \| = \sqrt{\tau_x^2 + \tau_y^2}$ [MPa]');





import bmcs_utils.api as bu
import traits.api as tr
import numpy as np
from scipy import interpolate
import bmcs_utils.api as bu

class TimeFunction(bu.InteractiveModel):

    name = 'Time function'
    n_i = bu.Int(0,input=True,desc='number of data points between data points')
    '''Number of steps between two values
    '''

    values = tr.Array(np.float_, value = [0,1])
    '''Values of the time function.
    '''

    ipw_view = bu.View(
        bu.Item('n_i', latex=r'n_i', minmax=(0,100))
    )

    t_step = tr.Property(tr.Float, depends_on='+input')
    '''Step size.
    '''
    @tr.cached_property
    def _get_t_step(self):
        n_values = len(self.values)
        return (self.n_i+1) * n_values

    tf = tr.Property(depends_on='data_points')
    def _get_tf(self):
        n_values = len(self.values)
        t_values = np.linspace(0,1,n_values)
        return interpolate.interp1d(t_values, self.values)

    def __call__(self, arg):
        return self.tf(arg)

    def update_plot(self, ax):
        n_values = len(self.values)
        t = np.linspace(0,1,(self.n_i)*(n_values-1) + n_values)
        v = self.tf(t)
        ax.plot(t, v, '-o')

if __name__ == '__main__':
    eps = [0, -0.01]
    eps = [0, -0.002745, -0.0016, -0.0047, -0.0025, -0.0065, -0.003, -0.0085, -0.004, -0.01]
    tf = TimeFunction(n_i=0, values=eps)
    import matplotlib.pyplot as plt
    fig, ax = plt.subplots(1,1)
    tf.plot(ax)
    plt.show()




# In[39]:

import numpy as np


# **Loop over the time increments** for a single material point. This loop emulates the  finite-element spatial integration or a lattice-assembly algorithm.

# In[42]:


def get_response(s_max=[3, 0, 0], n_steps=10, k_max=20, get_load_fn=lambda t: t, **kw):
    Eps_n1 = np.zeros((len(Eps),), dtype=np.float_)
    Sig_record = []
    Eps_record = []
    iter_record = []
    t_arr = np.linspace(0, 1, n_steps + 1)
    s_x_max, s_y_max, w_max = s_max
    s_x_t = s_x_max * get_load_fn(t_arr) + 1e-9
    s_y_t = s_y_max * get_load_fn(t_arr) + 1e-9
    w_t = w_max * get_load_fn(t_arr) + 1e-9
    for s_x_n1, s_y_n1, w_n1 in zip(s_x_t, s_y_t, w_t):
        Eps_n1, Sig_n1, k = get_material_model(s_x_n1, s_y_n1, w_n1, Eps_n1, k_max, **kw)
        Sig_record.append(Sig_n1)
        Eps_record.append(Eps_n1)
        iter_record.append(k)
    Sig_arr = np.array(Sig_record, dtype=np.float_)
    Eps_arr = np.array(Eps_record, dtype=np.float_)
    iter_arr = np.array(iter_record, dtype=np.int_)
    return t_arr, s_x_t, s_y_t, w_t, Eps_arr, Sig_arr, iter_arr




#!/usr/bin/env python
# coding: utf-8

# # Damage-plasticity SLIDE 3.2
# This notebook is a work in progress on an abstract and general implementation of time integration algorithm for general damage-plasticity modes. It serves for the development of a package that can be configured by specifying the ingredients of thermodynamically based model
# 
#  - Vector of state variables $\boldsymbol{\mathcal{E}}$
#  - Vector of thermodynamic streses $\boldsymbol{\mathcal{S}}$
#  - Helmholtz free energy $\psi(\boldsymbol{\mathcal{E}})$
#  - Threshold on thermodynamical forces  $f(\boldsymbol{\mathcal{S}},\boldsymbol{\mathcal{E}})$ / Yield condition
#  - Flow potential $\varphi(\boldsymbol{\mathcal{S}},\boldsymbol{\mathcal{E}})$
# 
# as symbolic equations using the sympy package. The time-stepping algorithm gets generated automatically within the thermodynamically framework. The derived  evolution equations and return-mapping to the yield surface is performed using Newton scheme.  

# In[1]:

import sympy as sp
import matplotlib.pyplot as plt
import numpy as np

# **Code generation** The derivation is adopted for the purpose of code generation both in Python and C utilizing the `codegen` package provided in `sympy`. The expressions that are part of the time stepping algorithm are transformed to an executable code directly at the place where they are derived. At the end of the notebook the C code can be exported to external files and applied in external tools. 

# In[2]:


# This code is needed to lambdify expressions named with latex symbols
# it removes the backslashes and curly braces upon before code generation.
from sympy.utilities.codegen import codegen
import re
def _print_Symbol(self, expr):
    CodePrinter = sp.printing.codeprinter.CodePrinter
    name = super(CodePrinter, self)._print_Symbol(expr)
    return re.sub(r'[\{^}]','_',re.sub(r'[\\\{\}]', '', name))
sp.printing.codeprinter.CodePrinter._print_Symbol = _print_Symbol

def ccode(cfun_name, sp_expr, cfile):
    '''Generate c function cfun_name for expr and directive name cfile 
    '''
    return codegen((cfun_name, sp_expr), 'C89', cfile + '_' + cfun_name)


# ## TODO
#  - The current implementation uses the threshold function linearized along $\lambda$ to represent the consistency condition. 
#  - As the next step include the fully linearized set of evolution equations and threshold functions. This might be important for SLIDE 3.x, 2D and 3D problems (von Mises, Drucker-Prager). In the current one-dimensional problem it probably has no effect - this feature will be included in SLIDE-core notebook which should be able to generate all the other problems by configuring the thermodynamic inputs symbolically.  
#  - Put the derived get_methods into a class and let the predictor-corrector implementation just access them. With this, the iteration scheme becomes completely generic. Any kind of model can be implemented then using the thermodynamic framework.
#  - Verification - check to see if the parameters controlling interaction between tension and shear have an effect and for which range of values
#  - Think of more flat profile of the f function to reduce the number of iterations during return mapping
#  - Visualize the iteration within the yield potential space for elementary loading scenarios.
#  - Check the reset of the interactive app upon change of the material parameters. Sometimes it is not replotted directly.

# &nbsp;<font color='blue'>
# **Naming conventions:**
#  - Variables with trailing underscore (e.g. `f_`, or `Sig_`) denote `sympy` expressions. 
#  - Variables denoting `sympy` symbols (e.g. `Sig` have no underscore at the end and have an the name which is close to the mathematical symbol
#  - Mathematical symbols defined as string in `sp.symbols(r'\tau^{\pi}')` use `latex` syntax to introduce Greek symbols, super and subindexes. This makes the pretty printing of expression possible.
#  - In an implemented algorithm at the end of the notebook, the Python variables containing the numerical values of the material parameters $E_b$, $\tau_\mathrm{Y}$, etc. are denoted with a leading underscore `_E_b` and `_tau_Y` to avoid name collisions within the notebook
# </font>

# ## Material parameters

# In[3]:


E_s = sp.Symbol('E_s', real=True, nonnegative=True)
gamma_s = sp.Symbol('gamma_s', real=True, nonnegative=True)
K_s = sp.Symbol('K_s', real=True)
S_s = sp.Symbol('S_s', real=True)
r_s = sp.Symbol('r_s', real=True)
c_s = sp.Symbol('c_s', real=True)
bartau = sp.Symbol(r'\bar{\tau}', real=True, nonnegative=True)
#bartau = sp.Symbol(r'tau_Y', real=True, nonnegative=True)

E_w = sp.Symbol('E_w', real=True, nonnegative=True)
S_w = sp.Symbol('S_w', real=True)
r_w = sp.Symbol('r_w', real=True)
c_w = sp.Symbol('c_w', real=True)
eta = sp.Symbol('eta', real=True, nonnegative=True)


# In[5]:



# ## State variables

# In[6]:


s_x, s_y = sp.symbols('s_x, s_y', real=True)
omega_s = sp.Symbol('omega_s', real=True, nonnegative=True)
s_pi_x, s_pi_y = sp.symbols(r's^{\pi}_x, s^{\pi}_y', real=True)
#s_pi_x, s_pi_y = sp.symbols(r's_pi_x, s_pi_y', real=True)
alpha_x, alpha_y = sp.symbols('alpha_x, alpha_y', real=True)
z = sp.Symbol('z', real=True)

w = sp.symbols('w', real=True)
omega_w = sp.Symbol('omega_w', real=True, nonnegative=True)
w_pi = sp.symbols(r'w^{\pi}', real=True)
#w_pi = sp.symbols(r'w_pi', real=True)


# In[7]:


Eps = sp.Matrix([s_pi_x, s_pi_y, w_pi, z, alpha_x, alpha_y, omega_s, omega_w])
Eps.T


# ## Helmholtz free energy

# The starting point in the thermodynamical representation of a process is a potential function of time dependent state variables. To describe the evolution of the state correctly describing the energy dissipation of the system the gradient of the potential function with respect to the state variables provides the generalized forces. The forces are constrained to characterize specific material properties, e.g. strength, hardening.

# In[8]:


rho_psi_s_ = sp.Rational(1,2)* ( 
    (1-omega_s)*E_s*(s_x-s_pi_x)**2 + 
    (1-omega_s)*E_s*(s_y-s_pi_y)**2 + 
    K_s * z**2 + 
    gamma_s * alpha_x**2 +
    gamma_s * alpha_y**2
)

rho_psi_w_ = sp.Rational(1,2) * (1 - omega_w) * E_w * (w - w_pi)**2


# In[9]:


rho_psi_ = rho_psi_s_ + rho_psi_w_
rho_psi_


# ## Thermodynamic forces

# In[10]:


tau_x, tau_y = sp.symbols('tau_x, tau_y', real=True)
tau_pi_x, tau_pi_y = sp.symbols(r'\tau^\pi_x, \tau^\pi_y', real=True)
#tau_pi_x, tau_pi_y = sp.symbols(r'tau_pi_x, tau_pi_y', real=True)
X_x, X_y = sp.symbols('X_x, X_y', real=True)
Z = sp.Symbol('Z', real=True, nonnegative=True)
Y_s = sp.Symbol('Y_s', real=True)

sig = sp.symbols('\sigma', real=True)
sig_pi = sp.symbols(r'\sigma^\pi', real=True)
#sig_pi = sp.symbols(r'sigma_pi', real=True)
Y_w = sp.Symbol('Y_w', real=True)


# In[11]:


Sig = sp.Matrix([tau_pi_x, tau_pi_y, sig_pi, Z, X_x, X_y, Y_s, Y_w])
Sig.T


# The introduce the thermodynamic forces we have to differentiate Hemholtz free energy
# with respect to the kinematic state variables
# \begin{align}
# \frac{\partial \rho \psi }{\partial \boldsymbol{\mathcal{E}}}
# \end{align}

# In[12]:


d_rho_psi_ = sp.Matrix([rho_psi_.diff(eps) for eps in Eps])
d_rho_psi_


# To obtain consistent signs of the Helmholtz derivatives we define a sign switch operator so that all generalized forces are defined as positive for the respective conjugate state variable $\boldsymbol{\Upsilon}$. 

# In[13]:


Sig_signs = sp.diag(-1,-1,-1,1,1,1,-1,-1)


# The constitutive laws between generalized force and kinematic variables then read
# \begin{align}
# \boldsymbol{\mathcal{S}} = \boldsymbol{\Upsilon}\frac{\rho \psi}{\partial\boldsymbol{\mathcal{E}}} 
# \end{align}

# In[14]:


Sig_ = Sig_signs * d_rho_psi_
Sig_.T


# **Executable code for** $\boldsymbol{\mathcal{S}}(s,\boldsymbol{\mathcal{E}})$

# In[15]:


# To derive the time stepping procedure we will need also the matrix of derivatives of the generalized stresses $\boldsymbol{\mathcal{S}}$ with respect to the kinematic variables $\boldsymbol{\mathcal{E}}$ 
# \begin{align}
# \frac{\partial \boldsymbol{S}}{\partial \boldsymbol{E}}
# \end{align}

# In[16]:


dSig_dEps_ = sp.Matrix([ 
    Sig_.T.diff(eps) for eps in Eps 
] ).T
dSig_dEps_


# **Executable Python code generation** $\displaystyle \frac{\partial }{\partial \boldsymbol{\mathcal{E}}}  \boldsymbol{\mathcal{S}}(s,\boldsymbol{\mathcal{E}})$

# In[17]:



# ## Threshold function

# To keep the framework general for different stress norms and hardening definitions let us first introduce a general function for effective stress. Note that the observable stress $\tau$ is identical with the plastic stress $\tau_\pi$ due to the performed sign switch in the definition of the thermodynamic forces.

# In[18]:


tau_eff_x = sp.Function(r'\tau^{\mathrm{eff}}_x')(tau_pi_x, omega_s)
tau_eff_y = sp.Function(r'\tau^{\mathrm{eff}}_y')(tau_pi_y, omega_s)
sig_eff = sp.Function(r'\sigma_{\mathrm{eff}}')(sig_pi, omega_w)
# tau_eff_x = sp.Function(r'tau_eff_x')(tau_pi_x, omega_s)
# tau_eff_y = sp.Function(r'tau_eff_y')(tau_pi_y, omega_s)
# sig_eff = sp.Function(r'sigma_eff')(sig_pi, omega_w)
Q_x = sp.Function('Q_x')(tau_eff_x,X_x)
Q_y = sp.Function('Q_y')(tau_eff_y,X_y)


# The stress norm is defined using the stress offset $X$, i.e. the kinematic hardening stress representing the shift of the origin of the yield locus.  

# In[19]:


norm_Q = sp.sqrt(Q_x*Q_x + Q_y*Q_y)


# Let us now introduce the back stress $X$ by defining the substitution for $Q = \tau^\mathrm{eff} - X$

# In[20]:


subs_Q = {Q_x: tau_eff_x - X_x, Q_y: tau_eff_y - X_y}
subs_Q


# Further substitution rule introduces the effective stress as a function of damage as

# In[21]:


subs_tau_eff = {tau_eff_x: tau_pi_x / (1-omega_s), tau_eff_y: tau_pi_y / (1-omega_s), sig_eff: sig_pi / (1-omega_w)}
subs_tau_eff


# After substitutions the yield function reads

# **Smooth yield function**

# In[24]:


# In[25]:

from bmcs_matmod.slide.f_double_cap import FDoubleCap
import bmcs_utils.api as bu

fdc = FDoubleCap()
f_t = fdc.symb.f_t
f_c = fdc.symb.f_c
f_c0 = fdc.symb.f_c0
m = fdc.symb.m
x = fdc.symb.x
y = fdc.symb.y

f_solved_ = fdc.symb.f_solved
f_1 = f_solved_.subs({x:sig_eff, y:norm_Q})
f_2 = f_1.subs(subs_Q)
f_3 = f_2.subs(subs_tau_eff)
f_ = f_3.subs(fdc.symb.tau_bar, (bartau+Z))

# **Executable code generation** $f(\boldsymbol{\mathcal{E}}, \boldsymbol{\mathcal{S}})$
# 
# Note that this is a function of both the forces and kinematic state variables

# The derivative of $f$ required for time-stepping $\frac{\partial f}{\partial \boldsymbol{\mathcal{S}}}$ is obtained as

df_dSig_ = f_.diff(Sig)
ddf_dEps_ = f_.diff(Eps)

phi_s_ext = (1-omega_s)**c_s * (Y_s**2 / S_s + eta * (Y_s * Y_w) / S_s)
phi_w_ext = (1-omega_w)**c_w * (Y_w**2 / S_w + eta * (Y_s * Y_w) / S_w) 

# The flow potential $\varphi(\boldsymbol{\mathcal{E}}, \boldsymbol{\mathcal{S}})$ reads

phi_ = f_ + phi_s_ext + phi_w_ext
phi_

# and the corresponding directions of flow given as a product of the sign operator $\Upsilon$ and of the derivatives with respect to state variables
# $\boldsymbol{\Upsilon} \, \partial_{\boldsymbol{\mathcal{S}}} \varphi$
# This renders following flow direction vector
# \begin{align}
# \boldsymbol{\Phi} = - \Upsilon \frac{\partial \varphi}{\partial \boldsymbol{\mathcal{S}}} 
# \end{align}

Phi_ = -Sig_signs * phi_.diff(Sig)

class Slide23Expr(bu.SymbExpr):

    # control and state variables
    s_x, s_y, w, Eps, Sig = s_x, s_y, w, Eps, Sig

    # model parameters
    E_s = E_s
    gamma_s = gamma_s
    K_s = K_s
    S_s = S_s
    r_s = r_s
    c_s = c_s
    bartau = bartau
    E_w = E_w
    S_w = S_w
    r_w = r_w
    c_w = c_w
    f_t = f_t
    f_c = f_c
    f_c0 = f_c0
    m = m
    eta = eta

    symb_model_params = [
        'E_s', 'gamma_s', 'K_s', 'S_s', 'c_s', 'bartau',
        'E_w', 'S_w', 'c_w', 'm', 'f_t', 'f_c', 'f_c0', 'eta'
    ]

    # expressions
    Sig_ = Sig_.T
    dSig_dEps_ = dSig_dEps_
    f_ = f_
    df_dSig_ = df_dSig_
    ddf_dEps_ = ddf_dEps_
    phi_ = phi_
    Phi_ = Phi_

    # List of expressions for which the methods `get_`
    symb_expressions = [
        ('Sig_', ('s_x', 's_y', 'w', 'Eps')),
        ('dSig_dEps_', ('s_x', 's_y', 'w', 'Eps')),
        ('f_', ('Eps', 'Sig')),
        ('df_dSig_', ('Eps', 'Sig')),
        ('ddf_dEps_', ('Eps', 'Sig')),
        ('phi_', ('Eps', 'Sig')),
        ('Phi_', ('Eps', 'Sig')),
    ]

#get_Sig_C = ccode('get_Sig',Sig_,'SLIDE1_3')
#get_dSig_dEps_C = ccode('get_dSig_dEps', dSig_dEps_, 'SLIDE1_3')
#get_f_C = ccode('get_f', f_, 'SLIDE1_3')
#get_df_dSig_C = ccode('get_df_dSig', df_dSig_, 'SLIDE1_3')
#get_ddf_dEps_C = ccode('get_df_dEps', ddf_dEps_, 'SLIDE1_3')
#get_Phi_C = ccode('get_Phi', Phi_, 'SLIDE1_3')

import traits.api as tr

class ConvergenceError(Exception):
    """ Inappropriate argument value (of correct type). """

    def __init__(self, message, state):  # real signature unknown
        self.message = message
        self.state = state
        super().__init__(self.message)

    def __str__(self):
        return f'{self.message} for state {self.state}'

class Slide32(bu.InteractiveModel,bu.InjectSymbExpr):

    name = 'Slide 3.2'
    symb_class = Slide23Expr

    E_s = bu.Float(28000, MAT=True)
    gamma_s = bu.Float(10, MAT=True)
    K_s = bu.Float(8, MAT=True)
    S_s = bu.Float(1, MAT=True)
    c_s = bu.Float(1, MAT=True)
    bartau = bu.Float(28000, MAT=True)
    E_w = bu.Float(28000, MAT=True)
    S_w =bu.Float(1, MAT=True)
    c_w = bu.Float(1, MAT=True)
    m = bu.Float(0.1, MAT=True)
    f_t = bu.Float(3, MAT=True)
    f_c = bu.Float(30, MAT=True)
    f_c0 = bu.Float(20, MAT=True)
    eta = bu.Float(0.5, MAT=True)

    ipw_view = bu.View(
        bu.Item('E_s', minmax=(0.5, 100)),
        bu.Item('S_s', minmax=(.00001, 100)),
        bu.Item('c_s', minmax=( 0.0001, 10)),
        bu.Item('gamma_s', latex=r'\gamma_\mathrm{s}', minmax=(-20, 20)),
        bu.Item('K_s', minmax=(-20, 20)),
        bu.Item('bartau', latex=r'\bar{\tau}', minmax=(0.5, 20)),
        bu.Item('E_w', minmax=(0.5, 100)),
        bu.Item('S_w', minmax=(0.0001, 100)),
        bu.Item('c_w', minmax=(0.0001, 10)),
        bu.Item('m', minmax=(0.0001, 0.4)),
        bu.Item('f_t', minmax=(0.1, 10)),
        bu.Item('f_c', latex=r'f_\mathrm{c}', minmax=(1, 200)),
        bu.Item('f_c0', latex=r'f_\mathrm{c0}', minmax=(1, 100)),
        bu.Item('eta', minmax=(0, 1))
    )

    def get_f_df(self, s_x_n1, s_y_n1, w_n1, Eps_k):
        Sig_k = self.symb.get_Sig_(s_x_n1, s_y_n1, w_n1, Eps_k)[0]
        dSig_dEps_k = self.symb.get_dSig_dEps_(s_x_n1, s_y_n1, w_n1, Eps_k)
        f_k = np.array([self.symb.get_f_(Eps_k, Sig_k)])
        df_dSig_k = self.symb.get_df_dSig_(Eps_k, Sig_k)
        ddf_dEps_k = self.symb.get_ddf_dEps_(Eps_k, Sig_k)
        df_dEps_k = np.einsum(
            'ik,ji->jk', df_dSig_k, dSig_dEps_k) + ddf_dEps_k
        Phi_k = self.symb.get_Phi_(Eps_k, Sig_k)
        dEps_dlambda_k = Phi_k
        df_dlambda = np.einsum(
            'ki,kj->ij', df_dEps_k, dEps_dlambda_k)
        df_k = df_dlambda
        return f_k, df_k, Sig_k

    def get_Eps_k1(self, s_x_n1, s_y_n1, w_n1, Eps_n, lam_k, Eps_k):
        '''Evolution equations:
        The update of state variables
        for an updated $\lambda_k$ is performed using this procedure.
        '''

        Sig_k = self.symb.get_Sig_(s_x_n1, s_y_n1, w_n1, Eps_k)[0]
        Phi_k = self.symb.get_Phi_(Eps_k, Sig_k)
        Eps_k1 = Eps_n + lam_k * Phi_k[:, 0]
        return Eps_k1

    rtol = bu.Float(1e-3, ALG=True)
    '''Relative tolerance of the return mapping algorithm related 
    to the tensile strength
    '''

    def get_sig_n1(self, s_x_n1, s_y_n1, w_n1, Eps_n, k_max):
        '''Return mapping iteration:
        This function represents a user subroutine in a finite element
        code or in a lattice model. The input is $s_{n+1}$ and the state variables
        representing the state in the previous solved step $\boldsymbol{\mathcal{E}}_n$.
        The procedure returns the stresses and state variables of
        $\boldsymbol{\mathcal{S}}_{n+1}$ and $\boldsymbol{\mathcal{E}}_{n+1}$
        '''
        Eps_k = np.copy(Eps_n)
        lam_k = 0
        f_k, df_k, Sig_k = self.get_f_df(s_x_n1, s_y_n1, w_n1, Eps_k)
        f_k_norm = np.linalg.norm(f_k)
        f_k_trial = f_k[0]
        k = 0
        while k < k_max:
            if f_k_trial < 0 or f_k_norm < self.f_t * self.rtol:
                return Eps_k, Sig_k, k + 1
            dlam = np.linalg.solve(df_k, -f_k)
            lam_k += dlam
            Eps_k = self.get_Eps_k1(s_x_n1, s_y_n1, w_n1, Eps_n, lam_k, Eps_k)
            f_k, df_k, Sig_k = self.get_f_df(s_x_n1, s_y_n1, w_n1, Eps_k)
            f_k_norm = np.linalg.norm(f_k)
            k += 1
        else:
            raise ConvergenceError('no convergence for step', [s_x_n1, s_y_n1, w_n1])

    def plot_f_state(self, ax, Eps, Sig):
        lower = -self.f_c * 1.05
        upper = self.f_t + 0.05 * self.f_c
        lower_tau = -self.bartau * 2
        upper_tau = self.bartau * 2
        tau_x, tau_y, sig = Sig[:3]
        tau = np.sqrt(tau_x**2 + tau_y**2)
        sig_ts, tau_x_ts  = np.mgrid[lower:upper:201j,lower_tau:upper_tau:201j]
        Sig_ts = np.zeros((len(self.symb.Eps),) + tau_x_ts.shape)
        Eps_ts = np.zeros_like(Sig_ts)
        Sig_ts[0,...] = tau_x_ts
        Sig_ts[2,...] = sig_ts
        Sig_ts[3:,...] = Sig[3:,np.newaxis,np.newaxis]
        Eps_ts[...] = Eps[:,np.newaxis,np.newaxis]
        f_ts = np.array([self.symb.get_f_(Eps_ts, Sig_ts)])
        ax.set_title('threshold function');
        ax.contour(sig_ts, tau_x_ts, f_ts[0,...], levels=0)
        ax.plot(sig, tau, marker='H', color='red')
        ax.plot([lower, upper], [0, 0], color='black', lw=0.4)
        ax.plot([0, 0], [lower_tau, upper_tau], color='black', lw=0.4)


    def plot_f(self, ax):
        lower = -self.f_c * 1.05
        upper = self.f_t + 0.05 * self.f_c
        lower_tau = -self.bartau * 2
        upper_tau = self.bartau * 2
        sig_ts, tau_x_ts  = np.mgrid[lower:upper:201j,lower_tau:upper_tau:201j]
        Sig_ts = np.zeros((len(self.symb.Eps),) + tau_x_ts.shape)
        Sig_ts[0,:] = tau_x_ts
        Sig_ts[2,:] = sig_ts
        Eps_ts = np.zeros_like(Sig_ts)
        f_ts = np.array([self.symb.get_f_(Eps_ts, Sig_ts)])
        ax.set_title('threshold function');
        ax.contour(sig_ts, tau_x_ts, f_ts[0,...], levels=0)
        ax.plot([lower, upper], [0, 0], color='black', lw=0.4)
        ax.plot([0, 0], [lower_tau, upper_tau], color='black', lw=0.4)

    def update_plot(self, ax):
        self.plot_f(ax)


# # Time integration scheme

# ## Discrete yield condition
# In a continuous case we consistency condition to explicitly glue the state onto the yield surface 
# \begin{align}
# \dot{f}(\boldsymbol{\mathcal{S}}(s, \boldsymbol{\mathcal{E}(\lambda)}), \boldsymbol{\mathcal{E}(\lambda)} ) &= 0 \end{align}
# In discrete case, we relax this requirement. Indeed, by taking $f(s_{n+1}; \boldsymbol{\mathcal{E}_n}) $ as a first trial value we can obtain positive values.
# 
# &nbsp;<font color="green">
# **We allow for "trial" states which lie beyond the admissible domain $f \le 0$ during iteration. This allows us to construct a "return mapping" algorithm that iteratively approaches an admissible state on the yield surface.**</font>

# Given an inadmissible trial state $k$ with the yield condition $f_k > 0$, let us introduce a linearized approximation of its change along the plastic multiplier $\lambda$ around the state $k$. 
# \begin{align}
#  f_{k+1} &= f_{k} + \left. \frac{\partial f}{\partial \lambda} \right|_k \Delta \lambda
# \end{align}
# In this form, we can search for an admissible state $f_{n+1} = 0$ by iterating over $k$.
# Note that in initial iteration $k = 0$ the state from previous step is reused, i.e. 
# $f(s_{n+1}; \boldsymbol{\mathcal{E}_n}) $.

# In the linearized form, we can transform the yield condition to a recurrent formula
# \begin{align}
# \left. \frac{\mathrm{d} f}{\mathrm{d} \lambda}\right|_k \Delta \lambda &= -f_k,
# \hspace{1cm} f_k \rightarrow 0 \; \;\mathrm{for}\;\; k = 1\ldots\infty
# \end{align}
# This resembles the Newton method for iterative solution of a nonlinear equation. However, we need to consider the fact that the level of inadmissibility $f$ changes between iterations. 
# ![image.png](attachment:image.png)
# Note that the predictor is negative and $\Delta \lambda > 0$. In every step, the plastic multiplier is updated:
# \begin{align}
# \lambda_{k+1} &= \lambda_k + \Delta \lambda, \, \lambda_0 = 0 \nonumber \\ \nonumber
# \end{align}

# Two more questions must addressed to define a general numerical algorithm for plasticity:
# <font color="brown">
#  * **Update of state variables $\boldsymbol{\mathcal{E}}_{k+1}$ in each iteration**
#  * **Expression of the predictor $\mathrm{d} f / \mathrm{d} \lambda$ in terms of the state variables**
# </font>

# ## State update
# In every iteration step the state variables $\boldsymbol{\mathcal{E}}$ must be updated using the discrete evolution equations, i.e. 
# \begin{align}
# \boldsymbol{\mathcal{E}}_{k+1} &= 
# \boldsymbol{\mathcal{E}}_k + \lambda_{k+1} 
# \boldsymbol{\Phi}_k
# \label{eq:discrete_evolution}
# \end{align}
# Which is used in the calculation of the threshold function in the next step. Note that $\boldsymbol{\Phi}_k$ is evaluated in the state $k$ and not $k+1$.

# To reach an admissible state let us linearize the threshold function at an interim state $k$ as
# \begin{align}
# f_{k+1} = 
# f_k 
#  +
# \left.
# \frac
# {\partial f}
# {\partial \lambda}
# \right|_k
# \Delta \lambda
# \end{align}

# ## Predictor
# \begin{align}
# \left.
# \frac{\partial f}{\partial{\lambda}}  
# \right|_k 
# &=
# \left.
# \frac{\partial f}{\partial{\boldsymbol{ \mathcal{E}}}}  
# \right|_k 
# \left.
# \frac{\partial {\boldsymbol{ \mathcal{E}}}}{\partial \lambda}
# \right|_k =
# \left.
# \frac{\partial f}{\partial{\boldsymbol{ \mathcal{E}}}}  
# \right|_k 
# \boldsymbol{\Phi}_k \\
# \left.
# \frac{\partial f}{\partial{\boldsymbol{ \mathcal{E}}}}  
# \right|_k 
# &=
# \left. \frac{\partial f}{ \partial \boldsymbol{\mathcal{S}}}\right|_{k}
# \left. \frac{\partial \boldsymbol{\mathcal{S}}}{\partial \boldsymbol{\mathcal{E}}}\right|_{k}
# +
# \left. \frac{\partial^{\mathrm{dir}} f}{ \partial^{\mathrm{dir}} \boldsymbol{\mathcal{E}}}\right|_{k}
# \label{eq:df_dlambda}
# \end{align}

# **Remark 1:** The derivative $\partial^\mathrm{dir}$ denotes the direct derivative with respect to $\boldsymbol{\mathcal{E}}$

# **Remark 2:** Note that $\displaystyle \frac{\partial \boldsymbol{\mathcal{E}}}{\partial \lambda}$ in equation $\eqref{eq:df_dlambda}$ can be obtained from the evolution equations $\eqref{eq:discrete_evolution}$
# \begin{align}
# \boldsymbol{\mathcal{E}}_k = \boldsymbol{\mathcal{E}}_n + \lambda \, \boldsymbol{\Phi}_k\; \implies
# \left.
# \frac{\partial {\boldsymbol{ \mathcal{E}}}}{\partial \lambda}
# \right|_k = 
# \boldsymbol{\Phi}_k
# \end{align}

# Thus, by rewriting the linearized equation as a recurrence formula, the iteration algorithm is obtained
# \begin{align}
# &
# \left.
# \frac{\partial f}{\partial{\lambda}}  
# \right|_k 
# \Delta \lambda
# = - f^{(k)}\\
# & \lambda_{k+1} = \lambda_{k} + \Delta \lambda \\
# & \boldsymbol{\mathcal{E}}_{k+1} = \boldsymbol{\mathcal{E}}_{k} + 
#  \lambda_{k} \, 
# \frac{\partial {\boldsymbol{ \mathcal{E}}}}{\partial \lambda}
#  \\
# &k = k + 1
# \end{align}

# ## Implementation concept
# The gradient operators needed for the time-stepping scheme have been derived above and are now available for the implementation of the numerical algorithm both in `Python` and `C89` languages
# 
# <table style="width:50%">
# <tr>
# <th>Symbol</th>
# <th>Python</th>
# <th>C89</th>
# </tr>
# <tr>
# <td>$\mathcal{S}(s, \boldsymbol{\mathcal{E}}) $  
# </td>
# <td>get_Sig</td>
# <td>get_Sig_C</td>
# </tr>
# <tr>
# <td>$\partial_\boldsymbol{\mathcal{E}}  \boldsymbol{\mathcal{S}}(s, \boldsymbol{\mathcal{E}}) $</td>
# <td>get_dSig_dEps</td>
# <td>get_dSig_dEps_C</td>
# </tr>
# <tr>
# <td>$ f(\boldsymbol{\mathcal{S}}, \boldsymbol{\mathcal{E}})$</td>
# <td>get_f</td>
# <td>get_f_C</td>
# </tr>
# <tr>
# <td>$\partial_\boldsymbol{\mathcal{S}} f(\boldsymbol{\mathcal{S}}, \boldsymbol{\mathcal{E}}) $  
# </td>
# <td>get_df_dSig</td>
# <td>get_df_dSig_C</td>
# </tr>
# <tr>
# <td>$\partial_\boldsymbol{\mathcal{E}} f(\boldsymbol{\mathcal{S}}, \boldsymbol{\mathcal{E}}) $</td>
# <td>get_df_dEps</td>
# <td>get_df_dEps_C</td>
# </tr>
# <tr>
# <td>$\partial_\boldsymbol{\mathcal{S}} \varphi(\boldsymbol{\mathcal{S}}, \boldsymbol{\mathcal{E}}) $</td>
# <td>get_Phi</td>
# <td>get_Phi_C</td>
# </tr>
# </table>

# **Threshold and its derivatives:** To avoid repeated calculation of the same expressions, let us put the evaluation of $f$ and $\partial_\lambda f$ into a single procedure. The iteration loop can be constructed in such a way that the predictor for the next step is calculated along with the residuum. In case that the residuum is below the required tolerance, the overhead for an extra calculated derivative is negligible or, with some care, can be even reused in the next time step.  


# # Code generation
# See the docs for the code generation, the latexified  sympy symbols
# must be substituted such that they can act as standard C variable names.
# The issue with this substitution might have been partially fixed by the substitution
# code defined by this code.
#
# The methods
#  * `get_f_df()`, and
#  * `get_Eps_n1` must be rewritten in C.
#
# **How to transform `einsum` to C?**
#
# The state arrays `Eps` and `Sig` must be prepared by the callee, i.e. the function to be
# called from at a level of material point in a finite-element or lattice code.
# All the state variable matrices are flattened in the generated C code so that index access operators must be constructed with the correct convention, i.e.
# `i * n_row + j` or `i + n_col * j`.
#
# Let us consider the line
# ```Python
# df_dEps_k = np.einsum('ik,ji->jk', df_dSig_k, dSig_dEps_k) + ddf_dEps_k
# ````
#
# To transform this systematically into the C loop it is proposed define
# index directives specifying the size of the first dimension. Then, the
# indexes from the `einsum` call can be systematically transferred to the
# multi-loop running over the indexes and respecting their order. No more
# thinking needed. The index operators below are prepared for arrays with
# 2, 3, and 4 dimensions.
# ```C
# #define IJ(N_I,I,J) (N_I * J + I)
# #define IJK(N_I,N_J,I,J,K) (N_I * IJ(N_J,J,K) + I)
# #define IJKL(N_I,N_J,N_K,I,J,K,L) (N_I * IJK(N_J,N_K,J,K,L) + I)
# ````
# to perform the matrix multiplication
# ```C
# int k=0;
# for(int i=0;i<N_I;i++)
#     for(int j=0;j<N_J;j++)
#         df_dEps_1[IJ(N_J,j,k)] += df_dSig_k[IJ(N_I(i,k)] * dSig_dEps_k[IJ(N_J,j,i)]
#     df_dEps[IJ(N_I,i,k)] = df_dEps_1[IJ(N_J,j,k)] + ddSig_dEps(IJ(N_I,i,k)];
# ````
#
# The whole material model is represented by the material method
# that mimics the internal par of the `get_response` function defined above
#
#  1. Call `get_f_df()` to get the trial state
#  2. Start the return mapping iteration loop
#  3. If admissibility criterion fulfilled - admissible state found - return stress
#  4. Evaluate delta of plastic multiplier and update it
#  5. Update state variables using evolution equations `get_Eps_n1`
#  6. Evaluate the residuum using `get_f_df()` and continuum with point 3.
#

# In[ ]:


# C_code = get_Sig_C, get_dSig_dEps_C, get_f_C, get_df_dSig_C, get_ddf_dEps_C, get_Phi_C

# In[ ]:


import os
import os.path as osp

def codegen():
    code_dirname = 'sympy_codegen'
    code_fname = 'SLIDE_1_3_2D'

    home_dir = osp.expanduser('~')
    code_dir = osp.join(home_dir, code_dirname)
    if not osp.exists(code_dir):
        os.makedirs(code_dir)

    code_file = osp.join(code_dir, code_fname)

    print('code_file', code_file)
    h_file = code_file + '.h'
    c_file = code_file + '.c'

    h_f = open(h_file, 'w')
    c_f = open(c_file, 'w')

    if True:
        for function_C in C_code:
            h_f.write(function_C[1][1])
            c_f.write(function_C[0][1])
    h_f.close()
    c_f.close()


# material_params = dict(
#     E_s=1, gamma_s=5, K_s=5, S_s=0.6, c_s=1, bartau=1,
#     E_w=1, S_w=0.6, c_w = 1, m = 0.01, f_t=1, f_c=-20, f_c0=-10, eta=0.5
# )
#
# slide = Slide32(**material_params)
#
# Eps_1 = np.zeros((8,), dtype=np.float_) + 0.1
# Eps_1[0] = 10
# Eps_1[2] = 0
#
# f_val = slide.symb.get_f_(Eps_1, np.zeros_like(Eps_1))
# print(f_val)



import bmcs_utils.api as bu
import sympy as sp
import numpy as np

# # Support functions
# To run some examples, let us define some infrastructure including a more complex loading history and postprocessing

# ## Loading history
# This implementation uses the symbolic machinery which is not necessary a simpler data point based implementation with `numpy.interp1d` would be better ... later


class LoadFn(bu.InteractiveModel):

    def __call__(self,t):
        return t

class CyclicLoadFnExpr(bu.SymbExpr):
    pass

class CyclicLoadFn(LoadFn):

    t, theta = sp.symbols(r't, \theta')
    n_cycles = 5
    A = 2
    ups = np.array([((theta - 2 * cycle) * A + (1 - A), theta - 2 * cycle <= 1)
                    for cycle in range(n_cycles)])
    downs = np.array([((1 - (theta - (2 * cycle + 1))) * A + (1 - A), (theta - (2 * cycle + 1)) <= 1)
                      for cycle in range(n_cycles)])
    ups[0, 0] = theta
    updowns = np.einsum('ijk->jik', np.array([ups, downs])).reshape(-1, 2)
    load_fn = sp.Piecewise(*updowns).subs(theta, t * n_cycles)
    get_load_fn = sp.lambdify(t, load_fn, 'numpy')
    t_arr = np.linspace(0, 1, 600)
    plt.plot(t_arr, get_load_fn(t_arr));





'''
Created on 04.07.2019
@author: Abdul
plotting tool for microplane models
'''

import os

import matplotlib

import matplotlib.pyplot as plt
import numpy as np


class Micro2Dplot():

    def get_2Dviz(self, n_mp, eps_N_Emn, eps_T_Emn, omega_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn, Z_N_Emn, X_N_Emn, Y_N_Emn, \
           omega_T_Emn, z_T_Emn, alpha_T_Emn, eps_T_pi_Emn, sigma_T_Emn, Z_T_pi_Emn, X_T_pi_Emn, Y_T_pi_Emn, \
           Disip_omena_N_Emn, Disip_omena_T_Emn, Disip_eps_p_N_Emn, Disip_eps_p_T_Emn, Disip_iso_N_Emn, \
           Disip_iso_T_Emn, Disip_kin_N_Emn, Disip_kin_T_Emn):

        rads = np.arange(0, (2 * np.pi), (2 * np.pi) / n_mp)
        font = {'family': 'DejaVu Sans',
                'size': 18}

        matplotlib.rc('font', **font)
        A = np.array(range(len(eps_N_p_Emn[::2])))
        A= A[::10]

        plt.figure(figsize=(9, 3))
        plt.subplot(241, projection='polar')
        for i in A:
            plt.plot(rads, eps_N_Emn[i, :], 'k')
        plt.title(r'$\varepsilon_N$')

        plt.subplot(242, projection='polar')
        for i in A:
            plt.plot(rads, omega_N_Emn[i, :], 'g')
        plt.title(r'$\omega_N$')

        plt.subplot(243, projection='polar')
        for i in A:
            plt.plot(rads, eps_N_p_Emn[i, :], 'g')
        plt.ylim(-1.2 * np.max(np.abs(eps_N_p_Emn)),
                 0.8 * np.max(np.abs(eps_N_p_Emn)))
        plt.title(r'$\varepsilon^p_N$')

        plt.subplot(244, projection='polar')
        for i in A:
            plt.plot(rads, sigma_N_Emn[i, :], 'b')
        plt.ylim(-1.2 * np.max(np.abs(sigma_N_Emn)),
                 0.8 * np.max(np.abs(sigma_N_Emn)))
        plt.title(r'$\sigma_N$')

        plt.subplot(245, projection='polar')
        for i in A:
            plt.plot(rads, np.abs(eps_T_Emn[i, :]), 'k')
        plt.title(r'$\varepsilon_T$')

        plt.subplot(246, projection='polar')
        for i in A:
            plt.plot(rads, omega_T_Emn[i, :], 'r')
        plt.title(r'$\omega_T$')

        plt.subplot(247, projection='polar')
        for i in A:
            plt.plot(rads, np.abs(eps_T_pi_Emn[i, :]), 'r')
        plt.title(r'$\varepsilon^{\pi}_T$')

        plt.subplot(248, projection='polar')
        for i in A:
            plt.plot(rads, np.abs(sigma_T_Emn[i, :]), 'b')
        plt.title(r'$\sigma_T$')

        plt.show()



'''
Created on 29.03.2017

@author: abaktheer

Microplane Fatigue model 2D

(compressive plasticity (CP) + tensile damage (TD) 
+ cumulative damage sliding (CSD))

Using Jirasek homogenization approach [1999]
'''

from bmcs_matmod.matmod.mats_base import MATSBase
from bmcs_utils.api import InteractiveModel, Item, View
import numpy as np
import traits.api as tr


class MATSXDMplCSDEEQ(MATSBase, InteractiveModel):

    gamma_T = tr.Float(100000.,
                       label="Gamma",
                        desc=" Tangential Kinematic hardening modulus",
                        enter_set=True,
                        auto_set=False)

    K_T = tr.Float(10000.,
                label="K",
                desc="Tangential Isotropic harening",
                enter_set=True,
                auto_set=False)

    S_T = tr.Float(0.005,
                label="S",
                desc="Damage strength",
                enter_set=True,
                auto_set=False)

    r_T = tr.Float(9.,
                label="r",
                desc="Damage cumulation parameter",
                enter_set=True,
                auto_set=False)
    e_T = tr.Float(12.,
                label="e",
                desc="Damage cumulation parameter",
                enter_set=True,
                auto_set=False)

    c_T = tr.Float(4.6,
                label="c",
                desc="Damage cumulation parameter",
                enter_set=True,
                auto_set=False)

    tau_pi_bar = tr.Float(1.7,
                       label="Tau_bar",
                       desc="Reversibility limit",
                       enter_set=True,
                       auto_set=False)

    a = tr.Float(0.003,
              label="a",
              desc="Lateral pressure coefficient",
              enter_set=True,
              auto_set=False)

    ipw_view = View(
        Item('gamma_T', latex=r'\gamma_\mathrm{T}', minmax=(10,100000)),
        Item('K_T', latex=r'K_\mathrm{T}', minmax=(10, 10000)),
        Item('S_T', latex=r'S_\mathrm{T}', minmax=(0.001, 0.01)),
        Item('r_T', latex=r'r_\mathrm{T}', minmax=(1, 3)),
        Item('e_T', latex=r'e_\mathrm{T}', minmax=(1, 40)),
        Item('c_T', latex=r'c_\mathrm{T}', minmax=(1, 10)),
        Item('tau_pi_bar', latex=r'\bar{\tau}', minmax=(1, 10)),
        Item('a', latex=r'a', minmax=(0.001, 3)),
    )

    # -------------------------------------------
    # Normal_Tension constitutive law parameters (without cumulative normal strain)
    # -------------------------------------------
    Ad = tr.Float(100.0,
               label="a",
               desc="brittleness coefficient",
               enter_set=True,
               auto_set=False)

    eps_0 = tr.Float(0.00008,
                  label="a",
                  desc="threshold strain",
                  enter_set=True,
                  auto_set=False)

    # -----------------------------------------------
    # Normal_Compression constitutive law parameters
    # -----------------------------------------------
    K_N = tr.Float(10000.,
                label="K_N",
                desc=" Normal isotropic harening",
                enter_set=True,
                auto_set=False)

    gamma_N = tr.Float(5000.,
                    label="gamma_N",
                    desc="Normal kinematic hardening",
                    enter_set=True,
                    auto_set=False)

    sigma_0 = tr.Float(30.,
                    label="sigma_0",
                    desc="Yielding stress",
                    enter_set=True,
                    auto_set=False)

    # -------------------------------------------------------------------------
    # Cached elasticity tensors
    # -------------------------------------------------------------------------

    E = tr.Float(35e+3,
                 label="E",
                 desc="Young's Modulus",
                 auto_set=False,
                 input=True)

    nu = tr.Float(0.2,
                  label='nu',
                  desc="Poison ratio",
                  auto_set=False,
                  input=True)

    def _get_lame_params(self):
        # la = self.E * self.nu / ((1. + self.nu) * (1. - 2. * self.nu))
        # # second Lame parameter (shear modulus)
        # mu = self.E / (2. + 2. * self.nu)
        la = self.E * self.nu / ((1. + self.nu) * (1. - self.nu))
        mu = self.E / (2. + 2. * self.nu)
        return la, mu

    D_abef = tr.Property(tr.Array, depends_on='+input')

    @tr.cached_property
    def _get_D_abef(self):
        la = self._get_lame_params()[0]
        mu = self._get_lame_params()[1]
        delta = np.identity(2)
        D_abef = (np.einsum(',ij,kl->ijkl', la, delta, delta) +
                  np.einsum(',ik,jl->ijkl', mu, delta, delta) +
                  np.einsum(',il,jk->ijkl', mu, delta, delta))
        return D_abef

    @tr.cached_property
    def _get_state_var_shapes(self):
        return {'w_N_Emn': (self.n_mp,),
                'z_N_Emn': (self.n_mp,),
                'alpha_N_Emn': (self.n_mp,),
                'r_N_Emn': (self.n_mp,),
                'eps_N_p_Emn': (self.n_mp,),
                'sigma_N_Emn': (self.n_mp,),
                'w_T_Emn': (self.n_mp,),
                'z_T_Emn': (self.n_mp,),
                'alpha_T_Emna': (self.n_mp, 2),
                'eps_T_pi_Emna': (self.n_mp, 2),
                }

    #--------------------------------------------------------------
    # microplane constitutive law (normal behavior CP + TD)
    # (without cumulative normal strain for fatigue under tension)
    #--------------------------------------------------------------
    def get_normal_law(self, eps_N_Emn, omega_N_Emn, z_N_Emn,
                       alpha_N_Emn, r_N_Emn, eps_N_p_Emn, eps_aux):

        eps_N_Aux = self._get_e_N_Emn_2(eps_aux)

        E_N = self.E / (1.0 - 2.0 * self.nu)

        # E_N = self.E * (1.0 + 2.0 *  self.nu) / (1.0 - self.nu**2)


        # When deciding if a microplane is in tensile or compression, we define a strain boundary such that that
        # sigmaN <= 0 if eps_N < 0, avoiding entering in the quadrant of compressive strains and traction

        sigma_trial = E_N * (eps_N_Emn - eps_N_p_Emn)
        pos1 = [(eps_N_Emn < -1e-6) & (sigma_trial > 1e-6)] # looking for microplanes violating strain boundary
        sigma_trial [pos1[0]] = 0
        pos = eps_N_Emn > 1e-6                            # microplanes under traction
        pos2 = eps_N_Emn < -1e-6                          # microplanes under compression
        H = 1.0 * pos
        H2 = 1.0 * pos2

        # thermo forces
        sigma_N_Emn_tilde = E_N * (eps_N_Emn - eps_N_p_Emn)
        sigma_N_Emn_tilde[pos1[0]] = 0                      # imposing strain boundary

        Z = self.K_N * z_N_Emn
        X = self.gamma_N * alpha_N_Emn * H2
        h = (self.sigma_0 + Z) * H2

        f_trial = (abs(sigma_N_Emn_tilde - X) - h) * H2

        # threshold plasticity

        thres_1 = f_trial > 1e-6

        delta_lamda = f_trial / \
            (E_N / (1 - omega_N_Emn) + abs(self.K_N) + self.gamma_N) * thres_1
        eps_N_p_Emn = eps_N_p_Emn + delta_lamda * \
            np.sign(sigma_N_Emn_tilde - X)
        z_N_Emn = z_N_Emn + delta_lamda
        alpha_N_Emn = alpha_N_Emn + delta_lamda * \
            np.sign(sigma_N_Emn_tilde - X)

        def R_N(r_N_Emn): return (1.0 / self.Ad) * (-r_N_Emn / (1.0 + r_N_Emn))

        Y_N = 0.5 * H * E_N * (eps_N_Emn - eps_N_p_Emn) ** 2.0
        Y_0 = 0.5 * E_N * self.eps_0 ** 2.0

        f = (Y_N - (Y_0 + R_N(r_N_Emn)))

        # threshold damage

        thres_2 = f > 1e-6

        def f_w(Y): return 1.0 - 1.0 / (1.0 + self.Ad * (Y - Y_0))

        omega_N_Emn[f > 1e-6] = f_w(Y_N)[f > 1e-6]
        r_N_Emn[f > 1e-6] = -omega_N_Emn[f > 1e-6]

        sigma_N_Emn = (1.0 - H * omega_N_Emn) * E_N * (eps_N_Emn - eps_N_p_Emn)
        pos1 = [(eps_N_Emn < -1e-6) & (sigma_trial > 1e-6)] # looking for microplanes violating strain boundary
        sigma_N_Emn[pos1[0]] = 0


        Z = self.K_N * z_N_Emn
        X = self.gamma_N * alpha_N_Emn * H2

        return omega_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn, Z, X, Y_N

    #-------------------------------------------------------------------------
    # microplane constitutive law (Tangential CSD)-(Pressure sensitive cumulative damage)
    #-------------------------------------------------------------------------
    def get_tangential_law(self, eps_T_Emna, omega_T_Emn, z_T_Emn,
                           alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn):

        E_T = self.E / (1.0 + self.nu)

        # E_T = self.E * (1.0 - 4 * self.nu) / \
        #     ((1.0 + self.nu) * (1.0 - 2 * self.nu))

        # E_T = self.E * (1.0 - 3.0 *  self.nu) / (1.0 - self.nu**2)


        # thermo forces

        sig_pi_trial = E_T * (eps_T_Emna - eps_T_pi_Emna)

        Z = self.K_T * z_T_Emn
        X = self.gamma_T * alpha_T_Emna
        norm_1 = np.sqrt(
            np.einsum(
                '...na,...na->...n',
                (sig_pi_trial - X), (sig_pi_trial - X))
        )
        Y = 0.5 * E_T * \
            np.einsum(
                '...na,...na->...n',
                (eps_T_Emna - eps_T_pi_Emna),
                (eps_T_Emna - eps_T_pi_Emna))

        # threshold

        f = norm_1 - self.tau_pi_bar - \
            Z + self.a * sigma_N_Emn

        plas_1 = f > 1e-6
        elas_1 = f < 1e-6

        delta_lamda = f / \
            (E_T / (1.0 - omega_T_Emn) + self.gamma_T + self.K_T) * plas_1

        norm_2 = 1.0 * elas_1 + np.sqrt(
            np.einsum(
                '...na,...na->...n',
                (sig_pi_trial - X), (sig_pi_trial - X))) * plas_1

        eps_T_pi_Emna[..., 0] = eps_T_pi_Emna[..., 0] + plas_1 * delta_lamda * \
            ((sig_pi_trial[..., 0] - X[..., 0]) /
             (1.0 - omega_T_Emn)) / norm_2
        eps_T_pi_Emna[..., 1] = eps_T_pi_Emna[..., 1] + plas_1 * delta_lamda * \
            ((sig_pi_trial[..., 1] - X[..., 1]) /
             (1.0 - omega_T_Emn)) / norm_2

        omega_T_Emn += ((1 - omega_T_Emn) ** self.c_T) * \
            (delta_lamda * (Y / self.S_T) ** self.r_T) * \
            (self.tau_pi_bar / (self.tau_pi_bar + self.a * sigma_N_Emn)) ** self.e_T

        alpha_T_Emna[..., 0] = alpha_T_Emna[..., 0] + plas_1 * delta_lamda * \
            (sig_pi_trial[..., 0] - X[..., 0]) / norm_2
        alpha_T_Emna[..., 1] = alpha_T_Emna[..., 1] + plas_1 * delta_lamda * \
            (sig_pi_trial[..., 1] - X[..., 1]) / norm_2

        z_T_Emn = z_T_Emn + delta_lamda

        sigma_T_Emna = np.einsum(
            '...n,...na->...na', (1 - omega_T_Emn), E_T * (eps_T_Emna - eps_T_pi_Emna))

        Z = self.K_T * z_T_Emn
        X = self.gamma_T * alpha_T_Emna
        Y = 0.5 * E_T * \
            np.einsum(
                '...na,...na->...n',
                (eps_T_Emna - eps_T_pi_Emna),
                (eps_T_Emna - eps_T_pi_Emna))

        return omega_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_T_Emna, Z, X, Y

#     #-------------------------------------------------------------------------
#     # MICROPLANE-Kinematic constraints
#     #-------------------------------------------------------------------------

    #-------------------------------------------------

    # get the operator of the microplane normals
    _MPNN = tr.Property(depends_on='n_mp')

    @tr.cached_property
    def _get__MPNN(self):
        print(self._MPN)
        MPNN_nij = np.einsum('ni,nj->nij', self._MPN, self._MPN)
        return MPNN_nij

    # get the third order tangential tensor (operator) for each microplane
    _MPTT = tr.Property(depends_on='n_mp')

    @tr.cached_property
    def _get__MPTT(self):
        delta = np.identity(2)
        MPTT_nijr = 0.5 * (
            np.einsum('ni,jr -> nijr', self._MPN, delta) +
            np.einsum('nj,ir -> njir', self._MPN, delta) - 2 *
            np.einsum('ni,nj,nr -> nijr', self._MPN, self._MPN, self._MPN)
        )
        return MPTT_nijr

    def _get_e_N_Emn_2(self, eps_Emab):
        # get the normal strain array for each microplane
        return np.einsum('nij,...ij->...n', self._MPNN, eps_Emab)

    def _get_e_T_Emnar_2(self, eps_Emab):
        # get the tangential strain vector array for each microplane
        MPTT_ijr = self._get__MPTT()
        return np.einsum('nija,...ij->...na', MPTT_ijr, eps_Emab)

    #---------------------------------------------------------------------
    # Extra homogenization of damage tensor in case of two damage parameters
    # Returns the 4th order damage tensor 'beta4' using (ref. [Baz99], Eq.(63))
    #---------------------------------------------------------------------

    def _get_beta_Emabcd_2(self, eps_Emab, omega_N_Emn, z_N_Emn,
                           alpha_N_Emn, r_N_Emn, eps_N_p_Emn, omega_T_Emn, z_T_Emn,
                           alpha_T_Emna, eps_T_pi_Emna, eps_aux):

        # Returns the 4th order damage tensor 'beta4' using
        #(cf. [Baz99], Eq.(63))

        eps_N_Emn = self._get_e_N_Emn_2(eps_Emab)
        eps_T_Emna = self._get_e_T_Emnar_2(eps_Emab)

        omega_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn, Z_n, X_n, Y_n = self.get_normal_law(
            eps_N_Emn, omega_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, eps_aux)

        omega_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_T_Emna, Z_T, X_T, Y_T = self.get_tangential_law(
            eps_T_Emna, omega_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn)

        delta = np.identity(2)
        beta_N = np.sqrt(1. - omega_N_Emn)
        beta_T = np.sqrt(1. - omega_T_Emn)

        beta_ijkl = np.einsum('n, ...n,ni, nj, nk, nl -> ...ijkl', self._MPW, beta_N, self._MPN, self._MPN, self._MPN, self._MPN) + \
            0.25 * (np.einsum('n, ...n,ni, nk, jl -> ...ijkl', self._MPW, beta_T, self._MPN, self._MPN, delta) +
                    np.einsum('n, ...n,ni, nl, jk -> ...ijkl', self._MPW, beta_T, self._MPN, self._MPN, delta) +
                    np.einsum('n, ...n,nj, nk, il -> ...ijkl', self._MPW, beta_T, self._MPN, self._MPN, delta) +
                    np.einsum('n, ...n,nj, nl, ik -> ...ijkl', self._MPW, beta_T, self._MPN, self._MPN, delta) -
                    4.0 * np.einsum('n, ...n, ni, nj, nk, nl -> ...ijkl', self._MPW, beta_T, self._MPN, self._MPN, self._MPN, self._MPN))

        return beta_ijkl
    #-----------------------------------------------------------
    # Integration of the (inelastic) strains for each microplane
    #-----------------------------------------------------------

    def _get_eps_p_Emab(self, eps_Emab, omega_N_Emn, z_N_Emn,
                        alpha_N_Emn, r_N_Emn, eps_N_p_Emn,
                        omega_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn, eps_aux):

        eps_N_Emn = self._get_e_N_Emn_2(eps_Emab)
        eps_T_Emna = self._get_e_T_Emnar_2(eps_Emab)

        # plastic normal strains
        omegaN, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn, Z_n, X_n, Y_n = self.get_normal_law(
            eps_N_Emn, omega_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, eps_aux)

        # sliding tangential strains
        omega_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_T_Emna, Z_T, X_T, Y_T = self.get_tangential_law(
            eps_T_Emna, omega_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn)

        delta = np.identity(2)

        # 2-nd order plastic (inelastic) tensor
        eps_p_Emab = (
            np.einsum('n,...n,na,nb->...ab',
                      self._MPW, eps_N_p_Emn, self._MPN, self._MPN) +
            0.5 * (
                np.einsum('n,...nf,na,fb->...ab',
                          self._MPW, eps_T_pi_Emna, self._MPN, delta) +
                np.einsum('n,...nf,nb,fa->...ab', self._MPW,
                          eps_T_pi_Emna, self._MPN, delta)
            )
        )

        return eps_p_Emab

    def get_corr_pred(self, eps_Emab, t_n1, int_var, eps_aux, F):
        '''Evaluation - get the corrector and predictor
        '''

        # Definition internal variables / forces per column:  1) damage N, 2)iso N, 3)kin N, 4)consolidation N, 5) eps p N,
        # 6) sigma N, 7) iso F N, 8) kin F N, 9) energy release N, 10) damage T, 11) iso T, 12-13) kin T, 14-15) eps p T,
        # 16-17) sigma T, 18) iso F T, 19-20) kin F T, 21) energy release T

        # Corrector predictor computation.

        #------------------------------------------------------------------
        # Damage tensor (4th order) using product- or sum-type symmetrization:
        #------------------------------------------------------------------


        eps_N_Emn = self._get_e_N_Emn_2(eps_Emab)
        eps_T_Emna = self._get_e_T_Emnar_2(eps_Emab)

        omega_N_Emn = int_var[:, 0]
        z_N_Emn = int_var[:, 1]
        alpha_N_Emn = int_var[:, 2]
        r_N_Emn = int_var[:, 3]
        eps_N_p_Emn = int_var[:, 4]
        sigma_N_Emn = int_var[:, 5]

        omega_T_Emn = int_var[:, 9]
        z_T_Emn = int_var[:, 10]
        alpha_T_Emna = int_var[:, 11:13]
        eps_T_pi_Emna = int_var[:, 13:15]


        beta_Emabcd = self._get_beta_Emabcd_2(
            eps_Emab, omega_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, omega_T_Emn, z_T_Emn,
                           alpha_T_Emna, eps_T_pi_Emna, eps_aux
        )

        #------------------------------------------------------------------
        # Damaged stiffness tensor calculated based on the damage tensor beta4:
        #------------------------------------------------------------------

        D_Emabcd = np.einsum(
            '...ijab, abef, ...cdef->...ijcd', beta_Emabcd, self.D_abef, beta_Emabcd)

        #----------------------------------------------------------------------
        # Return stresses (corrector) and damaged secant stiffness matrix (predictor)
        #----------------------------------------------------------------------
        # plastic strain tensor
        eps_p_Emab = self._get_eps_p_Emab(
            eps_Emab, omega_N_Emn, z_N_Emn,
            alpha_N_Emn, r_N_Emn, eps_N_p_Emn,
            omega_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn, eps_aux)

        # elastic strain tensor
        eps_e_Emab = eps_Emab - eps_p_Emab

        # calculation of the stress tensor
        sig_Emab = np.einsum('...abcd,...cd->...ab', D_Emabcd, eps_e_Emab)

        return D_Emabcd, sig_Emab, eps_p_Emab


class MATS2DMplCSDEEQ(MATSXDMplCSDEEQ):
    """Two dimensional version of the MS1 model
    """

    n_mp = tr.Constant(360)
    '''Number of microplanes
    '''

    #-----------------------------------------------
    # get the normal vectors of the microplanes
    #-----------------------------------------------
    _MPN = tr.Property(depends_on='n_mp')

    @tr.cached_property
    def _get__MPN(self):
        # microplane normals:
        alpha_list = np.linspace(0, 2 * np.pi, self.n_mp)

        MPN = np.array([[np.cos(alpha), np.sin(alpha)]
                        for alpha in alpha_list])

        return MPN

    #-------------------------------------
    # get the weights of the microplanes
    #-------------------------------------
    _MPW = tr.Property(depends_on='n_mp')

    @tr.cached_property
    def _get__MPW(self):
        MPW = np.ones(self.n_mp) / self.n_mp * 2

        return MPW

    #--------------------------------------------------------
    # return the state variables (Damage , inelastic strains)
    #--------------------------------------------------------
    def _get_state_variables(self, eps_Emab,
                             int_var, eps_aux):

        e_N_arr = self._get_e_N_Emn_2(eps_Emab)
        e_T_vct_arr = self._get_e_T_Emnar_2(eps_Emab)

        omega_N_Emn = int_var[:, 0]
        z_N_Emn = int_var[:, 1]
        alpha_N_Emn = int_var[:, 2]
        r_N_Emn = int_var[:, 3]
        eps_N_p_Emn = int_var[:, 4]
        sigma_N_Emn = int_var[:, 5]

        omega_T_Emn = int_var[:, 9]
        z_T_Emn = int_var[:, 10]
        alpha_T_Emna = int_var[:, 11:13]
        eps_T_pi_Emna = int_var[:, 13:15]

        omega_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn, Z_n, X_n, Y_n = self.get_normal_law(e_N_arr,  omega_N_Emn, z_N_Emn,
                                                                                                        alpha_N_Emn, r_N_Emn, eps_N_p_Emn, eps_aux)

        omega_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_T_Emna, Z_T, X_T, Y_T = self.get_tangential_law(e_T_vct_arr, omega_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn)

        # Definition internal variables / forces per column:  1) damage N, 2)iso N, 3)kin N, 4)consolidation N, 5) eps p N,
        # 6) sigma N, 7) iso F N, 8) kin F N, 9) energy release N, 10) damage T, 11) iso T, 12-13) kin T, 14-15) eps p T,
        # 16-17) sigma T, 18) iso F T, 19-20) kin F T, 21) energy release T

        int_var[:, 0] = omega_N_Emn
        int_var[:, 1] = z_N_Emn
        int_var[:, 2] = alpha_N_Emn
        int_var[:, 3] = r_N_Emn
        int_var[:, 4] = eps_N_p_Emn
        int_var[:, 5] = sigma_N_Emn
        int_var[:, 6] = Z_n
        int_var[:, 7] = X_n
        int_var[:, 8] = Y_n

        int_var[:, 9] = omega_T_Emn
        int_var[:, 10] = z_T_Emn
        int_var[:, 11:13] = alpha_T_Emna
        int_var[:, 13:15] = eps_T_pi_Emna
        int_var[:, 15:17] = sigma_T_Emna
        int_var[:, 17] = Z_T
        int_var[:, 18:20] = X_T
        int_var[:, 20] = Y_T


        return int_var




import bmcs_utils.api as bu
import traits.api as tr
from scipy.integrate import cumtrapz
import numpy as np

class EnergyDissipation(bu.InteractiveModel):
    name='Energy'

    colors = dict( # color associations
        stored_energy = 'darkgreen', # recoverable
        free_energy_kin = 'darkcyan', # freedom - sky
        free_energy_iso = 'darkslateblue', # freedom - sky
        plastic_diss_s = 'darkorange', # fire - heat
        plastic_diss_w = 'red', # fire - heat
        damage_diss_s = 'darkgray', # ruined
        damage_diss_w = 'black'  # ruined
    )
    slider_exp = tr.WeakRef(bu.InteractiveModel)

    t_arr = tr.DelegatesTo('slider_exp')
    Sig_arr = tr.DelegatesTo('slider_exp')
    Eps_arr = tr.DelegatesTo('slider_exp')
    s_x_t = tr.DelegatesTo('slider_exp')
    s_y_t = tr.DelegatesTo('slider_exp')
    w_t = tr.DelegatesTo('slider_exp')
    iter_t = tr.DelegatesTo('slider_exp')

    show_iter = bu.Bool(True)
    E_plastic_work = bu.Bool(True)
    E_iso_free_energy = bu.Bool(True)
    E_kin_free_energy = bu.Bool(True)
    E_app_plastic_diss = bu.Bool(True)
    E_damage_diss = bu.Bool(True)

    ipw_view = bu.View(
        bu.Item('show_iter'),
        bu.Item('E_damage_diss'),
        bu.Item('E_plastic_work'),
        bu.Item('E_iso_free_energy'),
        bu.Item('E_kin_free_energy'),
        bu.Item('E_app_plastic_diss'),
    )

    def plot_energy(self, ax, ax_i):

        W_arr = (
                cumtrapz(self.Sig_arr[:, 0], self.s_x_t, initial=0) +
                cumtrapz(self.Sig_arr[:, 1], self.s_y_t, initial=0) +
                cumtrapz(self.Sig_arr[:, 2], self.w_t, initial=0)
        )

        s_x_el_t = (self.s_x_t - self.Eps_arr[:, 0])
        s_y_el_t = (self.s_y_t - self.Eps_arr[:, 1])
        w_el_t = (self.w_t - self.Eps_arr[:, 2])
        U_arr = (
                self.Sig_arr[:, 0] * s_x_el_t / 2.0 +
                self.Sig_arr[:, 1] * s_y_el_t / 2.0 +
                self.Sig_arr[:, 2] * w_el_t / 2.0
        )
        G_arr = W_arr - U_arr
        ax.plot(self.t_arr, W_arr, lw=2, color='black', label=r'$W$ - Input work')
        ax.plot(self.t_arr, G_arr, '--', color='black', lw = 0.5, label=r'$W^\mathrm{inel}$ - Inelastic work')
        ax.fill_between(self.t_arr, W_arr, G_arr,
                        color=self.colors['stored_energy'], alpha=0.2)
        ax.set_xlabel('$t$ [-]');
        ax.set_ylabel(r'$E$ [Nmm]')
        ax.legend()

        E_i = cumtrapz(self.Sig_arr, self.Eps_arr, initial=0, axis=0)
        E_s_x_pi_, E_s_y_pi_, E_w_pi_, E_z_, E_alpha_x_, E_alpha_y_, E_omega_s_, E_omega_w_ = E_i.T

        E_plastic_work = E_s_x_pi_ + E_s_y_pi_ + E_w_pi_
        E_iso_free_energy = E_z_
        E_kin_free_energy = E_alpha_x_ + E_alpha_y_
        E_app_plastic_diss = E_plastic_work - E_iso_free_energy - E_kin_free_energy
        E_damage_diss = E_omega_s_ + E_omega_w_


        E_level = 0
        if self.E_damage_diss:
            ax.plot(self.t_arr, E_damage_diss + E_level, color='black', lw=1)
            ax.fill_between(self.t_arr, E_damage_diss + E_level, E_level, color='gray',
                            alpha=0.3);
            ax_i.plot(self.t_arr, E_damage_diss, color='gray', lw=2,
                      label=r'damage diss.: $Y\dot{\omega}$')
            ax.fill_between(self.t_arr, E_omega_w_ + E_level, E_level, color='black',
                            alpha=0.3);
            E_d_level = E_level + E_omega_w_
            ax.fill_between(self.t_arr, E_omega_s_ + E_d_level, E_d_level, color='blue',
                            alpha=0.3);
        E_level = E_damage_diss
        if self.E_plastic_work:
            ax.plot(self.t_arr, E_plastic_work + E_level, lw=1, color='red')
            ax.fill_between(self.t_arr, E_plastic_work + E_level, E_level, color='red', alpha=0.3)
            label = r'plastic work: $\sigma \dot{\varepsilon}^\pi$'
            ax_i.plot(self.t_arr, E_plastic_work, color='red', lw=2,label=label)
            ax.fill_between(self.t_arr, E_w_pi_ + E_level, E_level, color='black',
                            alpha=0.3);
            E_p_level = E_level + E_w_pi_
            ax.fill_between(self.t_arr, E_s_x_pi_ + E_s_y_pi_ + E_p_level, E_p_level, color='brown',
                            alpha=0.3);
        if self.E_app_plastic_diss:
            ax.plot(self.t_arr, E_app_plastic_diss + E_level, lw=1, color='magenta')
            ax.fill_between(self.t_arr, E_app_plastic_diss + E_level, E_level, color='magenta', alpha=0.3)
            label = r'apparent pl. diss.: $\sigma \dot{\varepsilon}^\pi - X\dot{\alpha} - Z\dot{z}$'
            ax_i.plot(self.t_arr, E_app_plastic_diss, color='magenta', lw=2, label=label)
        E_level += E_app_plastic_diss
        if self.E_iso_free_energy:
            ax.plot(self.t_arr, E_iso_free_energy + E_level, '-.', lw=1, color='black')
            ax.fill_between(self.t_arr, E_iso_free_energy + E_level, E_level, color='orange', alpha=0.3)
            ax_i.plot(self.t_arr, -E_iso_free_energy, '-.', color='orange', lw=2,
                      label=r'iso. diss.: $Z\dot{z}$')
        E_level += E_iso_free_energy
        if self.E_kin_free_energy:
            ax.plot(self.t_arr, E_kin_free_energy + E_level, '-.', color='black', lw=1)
            ax.fill_between(self.t_arr, E_kin_free_energy + E_level, E_level, color='blue', alpha=0.2);
            ax_i.plot(self.t_arr, -E_kin_free_energy, '-.', color='blue', lw=2,
                      label=r'free energy: $X\dot{\alpha}$')

        ax_i.legend()
        ax_i.set_xlabel('$t$ [-]');
        ax_i.set_ylabel(r'$E$ [Nmm]')

    @staticmethod
    def subplots(fig):
        ax_work, ax_energies = fig.subplots(1, 2)
        ax_iter = ax_work.twinx()
        return ax_work, ax_energies, ax_iter

    def update_plot(self, axes):
        ax_work, ax_energies, ax_iter = axes
        self.plot_energy(ax_work, ax_energies)
        if self.show_iter:
            ax_iter.plot(self.t_arr, self.iter_t)
            ax_iter.set_ylabel(r'$n_\mathrm{iter}$')

    def xsubplots(self, fig):
        ((ax1, ax2), (ax3, ax4)) = fig.subplots(2, 2, figsize=(10, 5), tight_layout=True)
        ax11 = ax1.twinx()
        ax22 = ax2.twinx()
        ax33 = ax3.twinx()
        ax44 = ax4.twinx()
        return ax1, ax11, ax2, ax22, ax3, ax33, ax4, ax44

    def xupdate_plot(self, axes):
        ax1, ax11, ax2, ax22, ax3, ax33, ax4, ax44 = axes
        self.get_response([6, 0, 0])
        # plot_Sig_Eps(s_x_t, Sig_arr, Eps_arr, iter_t, *axes)
        s_x_pi_, s_y_pi_, w_pi_, z_, alpha_x_, alpha_y_, omega_s_, omega_w_ = self.Eps_arr.T
        tau_x_pi_, tau_y_pi_, sig_pi_, Z_, X_x_, X_y_, Y_s_, Y_w_ = self.Sig_arr.T
        ax1.plot(self.w_t, sig_pi_, color='green')
        ax11.plot(self.s_x_t, tau_x_pi_, color='red')
        ax2.plot(self.w_t, omega_w_, color='green')
        ax22.plot(self.w_t, omega_s_, color='red')



'''
Created on 29.03.2017

@author: abaktheer

Microplane Fatigue model 3D

(compressive plasticity (CP) + tensile damage (TD) 
+ cumulative damage sliding (CSD))

Using Jirasek homogenization approach [1999]
'''

import numpy as np
import traits.api as tr
from bmcs_matmod.matmod.mats_base import MATSBase


class MATS3DMplCSDEEQ(MATSBase):
    concrete_type = tr.Int

    gamma_T = tr.Float(100000.,
                       label="Gamma",
                       desc=" Tangential Kinematic hardening modulus",
                       enter_set=True,
                       auto_set=False)

    K_T = tr.Float(10000.,
                   label="K",
                   desc="Tangential Isotropic harening",
                   enter_set=True,
                   auto_set=False)

    S_T = tr.Float(0.005,
                   label="S",
                   desc="Damage strength",
                   enter_set=True,
                   auto_set=False)

    r_T = tr.Float(9.,
                   label="r",
                   desc="Damage cumulation parameter",
                   enter_set=True,
                   auto_set=False)
    e_T = tr.Float(12.,
                   label="e",
                   desc="Damage cumulation parameter",
                   enter_set=True,
                   auto_set=False)

    c_T = tr.Float(4.6,
                   label="c",
                   desc="Damage cumulation parameter",
                   enter_set=True,
                   auto_set=False)

    tau_pi_bar = tr.Float(1.7,
                          label="Tau_bar",
                          desc="Reversibility limit",
                          enter_set=True,
                          auto_set=False)

    a = tr.Float(0.003,
                 label="a",
                 desc="Lateral pressure coefficient",
                 enter_set=True,
                 auto_set=False)

    # -------------------------------------------
    # Normal_Tension constitutive law parameters (without cumulative normal strain)
    # -------------------------------------------
    Ad = tr.Float(100.0,
                  label="a",
                  desc="brittleness coefficient",
                  enter_set=True,
                  auto_set=False)

    eps_0 = tr.Float(0.00008,
                     label="a",
                     desc="threshold strain",
                     enter_set=True,
                     auto_set=False)

    # -----------------------------------------------
    # Normal_Compression constitutive law parameters
    # -----------------------------------------------
    K_N = tr.Float(10000.,
                   label="K_N",
                   desc=" Normal isotropic harening",
                   enter_set=True,
                   auto_set=False)

    gamma_N = tr.Float(5000.,
                       label="gamma_N",
                       desc="Normal kinematic hardening",
                       enter_set=True,
                       auto_set=False)

    sigma_0 = tr.Float(30.,
                       label="sigma_0",
                       desc="Yielding stress",
                       enter_set=True,
                       auto_set=False)

    # -------------------------------------------------------------------------
    # Cached elasticity tensors
    # -------------------------------------------------------------------------

    E = tr.Float(35e+3,
                 label="E",
                 desc="Young's Modulus",
                 auto_set=False,
                 input=True)

    nu = tr.Float(0.2,
                  label='nu',
                  desc="Poison ratio",
                  auto_set=False,
                  input=True)

    def __init__(self, concrete_type):

        self.concrete_type = concrete_type

        if concrete_type == 0:  # # C40 MA

            # self.gamma_T = 60000
            #
            # self.K_T = 8000.0
            #
            # self.S_T = 0.0033
            #
            # self.r_T = 9.
            #
            # self.e_T = 12.
            #
            # self.c_T = 7.6
            #
            # self.tau_pi_bar = 2.2
            #
            # self.a = 0.003
            #
            # self.Ad = 800.0
            #
            # self.eps_0 = 0.00001
            #
            # self.K_N = 5000.
            #
            # self.gamma_N = 4000.
            #
            # self.sigma_0 = 30.

            self.gamma_T = 60000

            self.K_T = 8000.0

            self.S_T = 0.0032

            self.r_T = 9.

            self.e_T = 12.

            self.c_T = 7.6

            self.tau_pi_bar = 2.2

            self.a = 0.003

            self.Ad = 800.0

            self.eps_0 = 0.00001

            self.K_N = 4500.

            self.gamma_N = 10000.

            self.sigma_0 = 22.5

            # -------------------------------------------------------------------------
            # Cached elasticity tensors
            # -------------------------------------------------------------------------

            self.E = 35e+3

            self.nu = 0.2

        if concrete_type == 1:  # #   C80 MA

            self.gamma_T = 700000.

            self.K_T = 20000.0

            self.S_T = 0.0075

            self.r_T = 15.

            self.c_T = 7.

            self.e_T = 14.

            self.tau_pi_bar = 2.0

            self.a = 0.01

            self.Ad = 1700.0

            self.eps_0 = 0.0001

            self.K_N = 17000.

            self.gamma_N = 9000.

            self.sigma_0 = 45.

            self.E = 42e+3

            self.nu = 0.2

        if concrete_type == 2:  # PARAMETERS FOR C120 MA

            self.gamma_T = 2000000.

            self.K_T = 2200.0

            self.S_T = 0.011

            self.r_T = 17.5

            self.c_T = 12.

            self.e_T = 10.

            self.tau_pi_bar = 1.8

            self.a = 0.008

            self.Ad = 1000.0

            self.eps_0 = 0.0001

            self.K_N = 35000.

            self.gamma_N = 25000.

            self.sigma_0 = 80.

            self.E = 44e+3

            self.nu = 0.2

        if concrete_type == 3:  # TENSILE PARAMETRIC STUDY

            self.gamma_T = 70000.

            self.K_T = 10000.0

            self.S_T = 0.000000001

            self.r_T = 1.21

            self.e_T = 1.

            self.c_T = 1.85

            self.tau_pi_bar = 0.65
            self.a = 0.001

            self.Ad = 2500.0

            self.eps_0 = 0.000001

            self.K_N = 4000.

            self.gamma_N = 20000.

            self.sigma_0 = 10.

            self.E = 35e+3

            self.nu = 0.2

        # COMPRESSION
        if concrete_type == 4:
            self.gamma_T = 10000.

            self.K_T = 10000.0

            self.S_T = 0.000007

            self.r_T = 1.2

            self.e_T = 1.

            self.c_T = 1.25

            self.tau_pi_bar = 5.

            self.a = 0.001

            self.Ad = 1500.0

            self.eps_0 = 0.0001

            self.K_N = 10000.

            self.gamma_N = 10000.

            self.sigma_0 = 30.

            self.E = 35e+3

            self.nu = 0.2

        # BI-AXIAL ENVELOPE
        if concrete_type == 5:
            self.gamma_T = 10000.

            self.K_T = 10000.0

            self.S_T = 0.000007

            self.r_T = 1.2

            self.e_T = 1.

            self.c_T = 1.8

            self.tau_pi_bar = 5.

            self.a = 0.01

            self.Ad = 50000.0

            self.eps_0 = 0.00008

            self.K_N = 15000.

            self.gamma_N = 20000.

            self.sigma_0 = 30.

            self.E = 35e+3

            self.nu = 0.2

    # --------------------------------------------------------------
    # microplane constitutive law (normal behavior CP + TD)
    # --------------------------------------------------------------
    def get_normal_law(self, eps_N_Emn, omega_N_Emn, z_N_Emn,
                       alpha_N_Emn, r_N_Emn, eps_N_p_Emn, eps_aux):

        eps_N_Aux = self._get_e_N_Emn_2(eps_aux)

        E_N = self.E / (1.0 - 2.0 * self.nu)

        # When deciding if a microplane is in tensile or compression, we define a strain boundary such that that
        # sigmaN <= 0 if eps_N < 0, avoiding entering in the quadrant of compressive strains and traction

        sigma_trial = E_N * (eps_N_Emn - eps_N_p_Emn)
        pos1 = [(eps_N_Emn < -1e-6) & (sigma_trial > 1e-6)]  # looking for microplanes violating strain boundary
        sigma_trial[pos1[0]] = 0
        pos = sigma_trial > 1e-6  # microplanes under traction
        pos2 = sigma_trial < -1e-6  # microplanes under compression
        H = 1.0 * pos
        H2 = 1.0 * pos2

        # thermo forces
        sigma_N_Emn_tilde = E_N * (eps_N_Emn - eps_N_p_Emn)
        sigma_N_Emn_tilde[pos1[0]] = 0  # imposing strain boundary

        Z = self.K_N * z_N_Emn
        X = self.gamma_N * alpha_N_Emn * H2
        h = (self.sigma_0 + Z) * H2

        f_trial = (abs(sigma_N_Emn_tilde - X) - h) * H2

        # threshold plasticity

        thres_1 = f_trial > 1e-6

        delta_lamda = f_trial / \
                      (E_N / (1 - omega_N_Emn) + abs(self.K_N) + self.gamma_N) * thres_1
        eps_N_p_Emn = eps_N_p_Emn + delta_lamda * \
                      np.sign(sigma_N_Emn_tilde - X)
        z_N_Emn = z_N_Emn + delta_lamda
        alpha_N_Emn = alpha_N_Emn + delta_lamda * \
                      np.sign(sigma_N_Emn_tilde - X)

        def R_N(r_N_Emn): return (1.0 / self.Ad) * (-r_N_Emn / (1.0 + r_N_Emn))

        Y_N = 0.5 * H * E_N * (eps_N_Emn - eps_N_p_Emn) ** 2.0
        Y_0 = 0.5 * E_N * self.eps_0 ** 2.0

        f = (Y_N - (Y_0 + R_N(r_N_Emn)))

        # threshold damage

        thres_2 = f > 1e-6

        def f_w(Y): return 1.0 - 1.0 / (1.0 + self.Ad * (Y - Y_0))

        omega_N_Emn[f > 1e-6] = f_w(Y_N)[f > 1e-6]
        r_N_Emn[f > 1e-6] = -omega_N_Emn[f > 1e-6]

        sigma_N_Emn = (1.0 - H * omega_N_Emn) * E_N * (eps_N_Emn - eps_N_p_Emn)
        pos1 = [(eps_N_Emn < -1e-6) & (sigma_trial > 1e-6)]  # looking for microplanes violating strain boundary
        sigma_N_Emn[pos1[0]] = 0

        Z = self.K_N * z_N_Emn
        X = self.gamma_N * alpha_N_Emn * H2

        # pos2 = sigma_N_Emn * (eps_N_Emn - eps_N_Aux) > -1e-6
        # H2 = 1.0 * pos2
        # sigma_N_Emn = (1.0 - H * w_N_Emn) * E_N * (eps_N_Emn) * H2

        return omega_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn, Z, X, Y_N

    # -------------------------------------------------------------------------
    # microplane constitutive law (Tangential CSD)-(Pressure sensitive cumulative damage)
    # -------------------------------------------------------------------------
    def get_tangential_law(self, eps_T_Emna, omega_T_Emn, z_T_Emn,
                           alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn):

        E_T = self.E * (1.0 - 4 * self.nu) / \
              ((1.0 + self.nu) * (1.0 - 2 * self.nu))

        # thermo forces

        sig_pi_trial = E_T * (eps_T_Emna - eps_T_pi_Emna)

        Z = self.K_T * z_T_Emn
        X = self.gamma_T * alpha_T_Emna
        norm_1 = np.sqrt(
            np.einsum(
                '...na,...na->...n',
                (sig_pi_trial - X), (sig_pi_trial - X))
        )
        Y = 0.5 * E_T * \
            np.einsum(
                '...na,...na->...n',
                (eps_T_Emna - eps_T_pi_Emna),
                (eps_T_Emna - eps_T_pi_Emna))

        # threshold

        f = norm_1 - self.tau_pi_bar - \
            Z + self.a * sigma_N_Emn

        plas_1 = f > 1e-6
        elas_1 = f < 1e-6

        delta_lamda = f / \
                      (E_T / (1.0 - omega_T_Emn) + self.gamma_T + self.K_T) * plas_1

        norm_2 = 1.0 * elas_1 + np.sqrt(
            np.einsum(
                '...na,...na->...n',
                (sig_pi_trial - X), (sig_pi_trial - X))) * plas_1

        eps_T_pi_Emna[..., 0] = eps_T_pi_Emna[..., 0] + plas_1 * delta_lamda * \
                                ((sig_pi_trial[..., 0] - X[..., 0]) /
                                 (1.0 - omega_T_Emn)) / norm_2
        eps_T_pi_Emna[..., 1] = eps_T_pi_Emna[..., 1] + plas_1 * delta_lamda * \
                                ((sig_pi_trial[..., 1] - X[..., 1]) /
                                 (1.0 - omega_T_Emn)) / norm_2
        eps_T_pi_Emna[..., 2] = eps_T_pi_Emna[..., 2] + plas_1 * delta_lamda * \
                                ((sig_pi_trial[..., 2] - X[..., 2]) /
                                 (1.0 - omega_T_Emn)) / norm_2

        omega_T_Emn += ((1.0 - omega_T_Emn) ** self.c_T) * \
                       (delta_lamda * (Y / self.S_T) ** self.r_T)  # * \
        # (self.tau_pi_bar / (self.tau_pi_bar - self.a * sigma_kk / 3.0))

        alpha_T_Emna[..., 0] = alpha_T_Emna[..., 0] + plas_1 * delta_lamda * \
                               (sig_pi_trial[..., 0] - X[..., 0]) / norm_2
        alpha_T_Emna[..., 1] = alpha_T_Emna[..., 1] + plas_1 * delta_lamda * \
                               (sig_pi_trial[..., 1] - X[..., 1]) / norm_2
        alpha_T_Emna[..., 2] = alpha_T_Emna[..., 2] + plas_1 * delta_lamda * \
                               (sig_pi_trial[..., 2] - X[..., 2]) / norm_2
        z_T_Emn = z_T_Emn + delta_lamda

        sigma_T_Emna = np.einsum(
            '...n,...na->...na', (1 - omega_T_Emn), E_T * (eps_T_Emna - eps_T_pi_Emna))

        Z = self.K_T * z_T_Emn
        X = self.gamma_T * alpha_T_Emna
        Y = 0.5 * E_T * \
            np.einsum(
                '...na,...na->...n',
                (eps_T_Emna - eps_T_pi_Emna),
                (eps_T_Emna - eps_T_pi_Emna))

        return omega_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_T_Emna, Z, X, Y

    # -------------------------------------------------------------------------
    # MICROPLANE-Kinematic constraints
    # -------------------------------------------------------------------------
    # get the operator of the microplane normals
    _MPNN = tr.Property(depends_on='n_mp')

    @tr.cached_property
    def _get__MPNN(self):
        MPNN_nij = np.einsum('ni,nj->nij', self._MPN, self._MPN)
        return MPNN_nij

    # get the third order tangential tensor (operator) for each microplane
    _MPTT = tr.Property(depends_on='n_mp')

    @tr.cached_property
    def _get__MPTT(self):
        delta = np.identity(3)
        MPTT_nijr = 0.5 * (
                np.einsum('ni,jr -> nijr', self._MPN, delta) +
                np.einsum('nj,ir -> njir', self._MPN, delta) - 2 *
                np.einsum('ni,nj,nr -> nijr', self._MPN, self._MPN, self._MPN)
        )
        return MPTT_nijr

    def _get_e_N_Emn_2(self, eps_Emab):
        # get the normal strain array for each microplane
        return np.einsum('nij,...ij->...n', self._MPNN, eps_Emab)

    def _get_e_T_Emnar_2(self, eps_Emab):
        # get the tangential strain vector array for each microplane
        MPTT_ijr = self._get__MPTT()
        return np.einsum('nija,...ij->...na', MPTT_ijr, eps_Emab)

    # --------------------------------------------------------
    # return the state variables (Damage , inelastic strains)
    # --------------------------------------------------------
    def _get_state_variables(self, eps_Emab,
                             int_var, eps_aux):

        e_N_arr = self._get_e_N_Emn_2(eps_Emab)
        e_T_vct_arr = self._get_e_T_Emnar_2(eps_Emab)

        omega_N_Emn = int_var[:, 0]
        z_N_Emn = int_var[:, 1]
        alpha_N_Emn = int_var[:, 2]
        r_N_Emn = int_var[:, 3]
        eps_N_p_Emn = int_var[:, 4]
        sigma_N_Emn = int_var[:, 5]

        omega_T_Emn = int_var[:, 9]
        z_T_Emn = int_var[:, 10]
        alpha_T_Emna = int_var[:, 11:14]
        eps_T_pi_Emna = int_var[:, 14:17]

        omega_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn, Z_n, X_n, Y_n = self.get_normal_law(
            e_N_arr, omega_N_Emn, z_N_Emn,
            alpha_N_Emn, r_N_Emn, eps_N_p_Emn, eps_aux)

        omega_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_T_Emna, Z_T, X_T, Y_T = self.get_tangential_law(
            e_T_vct_arr, omega_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn)

        # Definition internal variables / forces per column:  1) damage N, 2)iso N, 3)kin N, 4) consolidation N, 5) eps p N,
        # 6) sigma N, 7) iso F N, 8) kin F N, 9) energy release N, 10) damage T, 11) iso T, 12-14) kin T, 15-17) eps p T,
        # 18-20) sigma T, 21) iso F T, 22-24) kin F T, 25) energy release T

        int_var[:, 0] = omega_N_Emn
        int_var[:, 1] = z_N_Emn
        int_var[:, 2] = alpha_N_Emn
        int_var[:, 3] = r_N_Emn
        int_var[:, 4] = eps_N_p_Emn
        int_var[:, 5] = sigma_N_Emn
        int_var[:, 6] = Z_n
        int_var[:, 7] = X_n
        int_var[:, 8] = Y_n

        int_var[:, 9] = omega_T_Emn
        int_var[:, 10] = z_T_Emn
        int_var[:, 11:14] = alpha_T_Emna
        int_var[:, 14:17] = eps_T_pi_Emna
        int_var[:, 17:20] = sigma_T_Emna
        int_var[:, 20] = Z_T
        int_var[:, 21:24] = X_T
        int_var[:, 24] = Y_T

        return int_var

        # ---------------------------------------------------------------------
        # Extra homogenization of damage tensor in case of two damage parameters
        # Returns the 4th order damage tensor 'beta4' using (ref. [Baz99], Eq.(63))
        # ---------------------------------------------------------------------

    def _get_beta_Emabcd_2(self, eps_Emab, omega_N_Emn, z_N_Emn,
                           alpha_N_Emn, r_N_Emn, eps_N_p_Emn, omega_T_Emn, z_T_Emn,
                           alpha_T_Emna, eps_T_pi_Emna, eps_aux):
        # Returns the 4th order damage tensor 'beta4' using
        # (cf. [Baz99], Eq.(63))

        eps_N_Emn = self._get_e_N_Emn_2(eps_Emab)
        eps_T_Emna = self._get_e_T_Emnar_2(eps_Emab)

        omega_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn, Z_n, X_n, Y_n = self.get_normal_law(
            eps_N_Emn, omega_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, eps_aux)

        omega_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_T_Emna, Z_T, X_T, Y_T = self.get_tangential_law(
            eps_T_Emna, omega_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn)

        delta = np.identity(3)
        beta_N = np.sqrt(1. - omega_N_Emn)
        beta_T = np.sqrt(1. - omega_T_Emn)

        beta_ijkl = np.einsum('n, ...n,ni, nj, nk, nl -> ...ijkl', self._MPW, beta_N, self._MPN, self._MPN,
                              self._MPN, self._MPN) + \
                    0.25 * (np.einsum('n, ...n,ni, nk, jl -> ...ijkl', self._MPW, beta_T, self._MPN, self._MPN,
                                      delta) +
                            np.einsum('n, ...n,ni, nl, jk -> ...ijkl', self._MPW, beta_T, self._MPN, self._MPN,
                                      delta) +
                            np.einsum('n, ...n,nj, nk, il -> ...ijkl', self._MPW, beta_T, self._MPN, self._MPN,
                                      delta) +
                            np.einsum('n, ...n,nj, nl, ik -> ...ijkl', self._MPW, beta_T, self._MPN, self._MPN,
                                      delta) -
                            4.0 * np.einsum('n, ...n, ni, nj, nk, nl -> ...ijkl', self._MPW, beta_T, self._MPN,
                                            self._MPN, self._MPN, self._MPN))

        return beta_ijkl

    # -----------------------------------------------------------
    # Integration of the (inelastic) strains for each microplane
    # -----------------------------------------------------------

    def _get_eps_p_Emab(self, eps_Emab, omega_N_Emn, z_N_Emn,
                        alpha_N_Emn, r_N_Emn, eps_N_p_Emn,
                        omega_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn, eps_aux):
        eps_N_Emn = self._get_e_N_Emn_2(eps_Emab)
        eps_T_Emna = self._get_e_T_Emnar_2(eps_Emab)

        # plastic normal strains
        omegaN, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn, Z_n, X_n, Y_n = self.get_normal_law(
            eps_N_Emn, omega_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, eps_aux)

        # sliding tangential strains
        omega_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_T_Emna, Z_T, X_T, Y_T = self.get_tangential_law(
            eps_T_Emna, omega_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn)

        delta = np.identity(3)

        # 2-nd order plastic (inelastic) tensor
        eps_p_Emab = (
                np.einsum('n,...n,na,nb->...ab',
                          self._MPW, eps_N_p_Emn, self._MPN, self._MPN) +
                0.5 * (
                        np.einsum('n,...nf,na,fb->...ab',
                                  self._MPW, eps_T_pi_Emna, self._MPN, delta) +
                        np.einsum('n,...nf,nb,fa->...ab', self._MPW,
                                  eps_T_pi_Emna, self._MPN, delta)
                )
        )

        return eps_p_Emab

    # -------------------------------------------------------------------------
    # Evaluation - get the corrector and predictor
    # -------------------------------------------------------------------------

    def get_corr_pred(self, eps_Emab, t_n1, int_var, eps_aux, F):
        # Definition internal variables / forces per column:  1) damage N, 2)iso N, 3)kin N, 4)consolidation N, 5) eps p N,
        # 6) sigma N, 7) iso F N, 8) kin F N, 9) energy release N, 10) damage T, 11) iso T, 12-13) kin T, 14-15) eps p T,
        # 16-17) sigma T, 18) iso F T, 19-20) kin F T, 21) energy release T

        # Corrector predictor computation.

        # ------------------------------------------------------------------
        # Damage tensor (4th order) using product- or sum-type symmetrization:
        # ------------------------------------------------------------------

        eps_N_Emn = self._get_e_N_Emn_2(eps_Emab)
        eps_T_Emna = self._get_e_T_Emnar_2(eps_Emab)

        omega_N_Emn = int_var[:, 0]
        z_N_Emn = int_var[:, 1]
        alpha_N_Emn = int_var[:, 2]
        r_N_Emn = int_var[:, 3]
        eps_N_p_Emn = int_var[:, 4]
        sigma_N_Emn = int_var[:, 5]

        omega_T_Emn = int_var[:, 9]
        z_T_Emn = int_var[:, 10]
        alpha_T_Emna = int_var[:, 11:14]
        eps_T_pi_Emna = int_var[:, 14:17]
        sigma_T_Emna = int_var[:, 17:20]

        beta_Emabcd = self._get_beta_Emabcd_2(
            eps_Emab, omega_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, omega_T_Emn, z_T_Emn,
            alpha_T_Emna, eps_T_pi_Emna, eps_aux
        )

        # ------------------------------------------------------------------
        # Damaged stiffness tensor calculated based on the damage tensor beta4:
        # ------------------------------------------------------------------

        D_Emabcd = np.einsum(
            '...ijab, abef, ...cdef->...ijcd', beta_Emabcd, self.D_abef, beta_Emabcd)

        # ----------------------------------------------------------------------
        # Return stresses (corrector) and damaged secant stiffness matrix (predictor)
        # ----------------------------------------------------------------------
        # plastic strain tensor
        eps_p_Emab = self._get_eps_p_Emab(
            eps_Emab, omega_N_Emn, z_N_Emn,
            alpha_N_Emn, r_N_Emn, eps_N_p_Emn,
            omega_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn, eps_aux)

        # elastic strain tensor
        eps_e_Emab = eps_Emab - eps_p_Emab

        delta = np.identity(3)
        # calculation of the stress tensor
        sig_Emab = np.einsum('...abcd,...cd->...ab', D_Emabcd, eps_e_Emab)
        sig_Emab_int = (
                np.einsum('n,...n,na,nb->...ab',
                          self._MPW, sigma_N_Emn, self._MPN, self._MPN) + \
                0.5 * np.einsum('n,...ne,na,eb->...ab',
                                self._MPW, sigma_T_Emna, self._MPN, delta) + \
                0.5 * np.einsum('n,...ne,nb,ea->...ab',
                                self._MPW, sigma_T_Emna, self._MPN, delta))

        return D_Emabcd, sig_Emab, eps_p_Emab, sig_Emab_int

    # -----------------------------------------------
    # number of microplanes - currently fixed for 3D
    # -----------------------------------------------
    n_mp = tr.Constant(28)

    # -----------------------------------------------
    # get the normal vectors of the microplanes
    # -----------------------------------------------
    _MPN = tr.Property(depends_on='n_mp')

    @tr.cached_property
    def _get__MPN(self):
        return np.array([[.577350259, .577350259, .577350259],
                         [.577350259, .577350259, -.577350259],
                         [.577350259, -.577350259, .577350259],
                         [.577350259, -.577350259, -.577350259],
                         [.935113132, .250562787, .250562787],
                         [.935113132, .250562787, -.250562787],
                         [.935113132, -.250562787, .250562787],
                         [.935113132, -.250562787, -.250562787],
                         [.250562787, .935113132, .250562787],
                         [.250562787, .935113132, -.250562787],
                         [.250562787, -.935113132, .250562787],
                         [.250562787, -.935113132, -.250562787],
                         [.250562787, .250562787, .935113132],
                         [.250562787, .250562787, -.935113132],
                         [.250562787, -.250562787, .935113132],
                         [.250562787, -.250562787, -.935113132],
                         [.186156720, .694746614, .694746614],
                         [.186156720, .694746614, -.694746614],
                         [.186156720, -.694746614, .694746614],
                         [.186156720, -.694746614, -.694746614],
                         [.694746614, .186156720, .694746614],
                         [.694746614, .186156720, -.694746614],
                         [.694746614, -.186156720, .694746614],
                         [.694746614, -.186156720, -.694746614],
                         [.694746614, .694746614, .186156720],
                         [.694746614, .694746614, -.186156720],
                         [.694746614, -.694746614, .186156720],
                         [.694746614, -.694746614, -.186156720]])

    # -------------------------------------
    # get the weights of the microplanes
    # -------------------------------------
    _MPW = tr.Property(depends_on='n_mp')

    @tr.cached_property
    def _get__MPW(self):
        return np.array([.0160714276, .0160714276, .0160714276, .0160714276, .0204744730,
                         .0204744730, .0204744730, .0204744730, .0204744730, .0204744730,
                         .0204744730, .0204744730, .0204744730, .0204744730, .0204744730,
                         .0204744730, .0158350505, .0158350505, .0158350505, .0158350505,
                         .0158350505, .0158350505, .0158350505, .0158350505, .0158350505,
                         .0158350505, .0158350505, .0158350505]) * 6.0

    # -------------------------------------------------------------------------
    # Cached elasticity tensors
    # -------------------------------------------------------------------------

    def _get_lame_params(self):
        la = self.E * self.nu / ((1. + self.nu) * (1. - 2. * self.nu))
        # second Lame parameter (shear modulus)
        mu = self.E / (2. + 2. * self.nu)
        return la, mu

    D_abef = tr.Property(tr.Array, depends_on='+input')

    @tr.cached_property
    def _get_D_abef(self):
        la = self._get_lame_params()[0]
        mu = self._get_lame_params()[1]
        delta = np.identity(3)
        D_abef = (np.einsum(',ij,kl->ijkl', la, delta, delta) +
                  np.einsum(',ik,jl->ijkl', mu, delta, delta) +
                  np.einsum(',il,jk->ijkl', mu, delta, delta))

        return D_abef
